"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: Quangtuan Kieu
>>> all rights reserved
________________________________
"""

from numpy import pi,sqrt,dot

import numpy as np
import matplotlib.pyplot as plt

from tqix.qobj import *
from tqix.utility import randunit,randnormal
from tqix.qtool import dotx
from tqix.backend import cdf
from tqix.qstate import add_random_noise,add_white_noise
from tqix.dsm.util import gtrace,gfide

__all__ = ['execu']

def execu(state,ncdf,nsamp,theta,m,st,pHist='False'):
   # to run Probe-controll Configuration 2 DSM
   dim  = state.shape[0]

   ntr = np.zeros(nsamp)
   nfi = np.zeros(nsamp)
   
   if typex(state)=='oper':
      for i in range(nsamp):
         #p = randunit()
         p = randnormal(0.5,st)
         while (p < 0) or (p > 1):
            p = randnormal(0.5,st)
         state_new = add_white_noise(state,p)
         post_state = get_post_state(dim,m,st)
         prob = get_prob(state_new,post_state,theta,dim)
         rho = get_rho(state_new,prob,ncdf,dim)
         restate = get_qurec(post_state,theta,rho,dim)

         ntr[i] = gtrace(restate,state)
         nfi[i] = gfide(restate,state)
   else:
      for i in range(nsamp):
         state_new = add_random_noise(state,m,st)
         post_state = get_post_state(dim,m,st)
         restate = get_pqurec(state_new,post_state,theta,ncdf,dim)

         ntr[i] = gtrace(restate,state)
         nfi[i] = gfide(restate,state)

   # if plot hist
   if pHist:
      plt.hist(nfi,bins=100)
      #plt.show()

   # take average
   atr = sum(ntr)/float(nsamp)
   afi = sum(nfi)/float(nsamp)
   etr,efi = 0.0,0.0

   for i in range(nsamp):
      etr += (ntr[i]-atr)**2
      efi += (nfi[i]-afi)**2

   etr /= float(nsamp)
   efi /= float(nsamp)

   return atr,np.sqrt(etr),afi,np.sqrt(efi)

"""
-------------------
some common modules 
-------------------
"""

# get post_state
def get_post_state(dim,m = 0.0,st = 0.01):
    """
    to generate a post-selected state

    >>> |c0> = 1/√dim ∑(1+detal_m)|m>

    Parameters:
    -----------
    dim : dimension
    detal_m : normal Distribusion 
    st: standard deviation (default 0.01)

    -----------
    Return 1 + delta_m
    """
    post_state = 1 + randnormal(m,st,dim)
    post_state /= np.linalg.norm(post_state)    
    return qx(post_state)

# get propability
def get_prob(state,post_state,theta,dim):
   # to get probability 

   if typex(state)=='oper':
      temp_prob = np.zeros((3,dim,dim),dtype=complex)
      rho = np.zeros((4,dim,dim),dtype=complex)
      prob = np.zeros((6,dim,dim),dtype=complex)
      for p in range(dim):
         for k in range(dim):
            # temp_prob_0
            temp_prob[0,p,k] = state[p,p]
            # temp_prob_1: sum_{m}
            for m in range(dim):
               temp_prob[1,p,k] += \
               state[p,m]*post_state[0,m]*post_state[0,p]\
               *np.exp(1j*2.0*pi*k*(m-p)/dim)
            # temp_prob_2: sum_{n,m}
            for n in range(dim):
               for m in range(dim):
                  temp_prob[2,p,k] += \
                  state[n,m]*post_state[0,n]*post_state[0,m]\
                  *np.exp(1j*2.0*pi*k*(m-n)/dim)\
                  *(post_state[0,p])**2

      #rho_00
      rho[0,:,:] =  (temp_prob[0,:,:]-2.0*\
                    np.real(temp_prob[1,:,:])\
                    +temp_prob[2,:,:])/2.0
      #rho_01
      rho[1,:,:] = (temp_prob[1,:,:]-temp_prob[2,:,:])/2.0

      #rho_10
      rho[2,:,:] = np.conj(rho[1,:,:])

      #rho_11
      rho[3,:,:] = (temp_prob[2,:,:])/2.0
      #probabilities
      prob[0,:,:] = (rho[0,:,:]+rho[1,:,:]\
                     +rho[2,:,:]+rho[3,:,:])/2.0 #Prob.+
      prob[1,:,:] = (rho[0,:,:]-rho[1,:,:]\
                     -rho[2,:,:]+rho[3,:,:])/2.0 #Prob.-
      prob[2,:,:] = (rho[0,:,:]+1j*rho[1,:,:]\
                     -1j*rho[2,:,:]+rho[3,:,:])/2.0 #Prob.L
      prob[3,:,:] = (rho[0,:,:]-1j*rho[1,:,:]\
                     +1j*rho[2,:,:]+rho[3,:,:])/2.0 #Prob.R
      prob[4,:,:] = rho[0,:,:] #Prob. 0
      prob[5,:,:] = rho[3,:,:] #Prob. 1
      return np.real(prob)
   else: #for pure state
      prob = np.zeros((6,dim),dtype=complex)

      rpsi = np.real(state) #ket
      ipsi = np.imag(state) #ket

      psi_til = abs(dotx(post_state,state)[0,0]) #scala
      psi_abs2 = abs(state)**2 #ket: calculate |\psi'_n|^2

      #calculate probabilities
      prob[0,:]= (psi_abs2[:,0] - \
                  2.0*post_state[0,:]*psi_til*rpsi[:,0]+\
                  post_state[0,:]**2*psi_til**2)/(2.0) #P(0)
      prob[1,:]= (post_state[0,:]**2*psi_til**2)/(2.0) #P(1)
      prob[2,:]= (psi_abs2[:,0])/(4.0) #P(+)
      prob[3,:]= (psi_abs2[:,0]-4.0*psi_til*post_state[0,:]*rpsi[:,0]+\
               4.0*post_state[0,:]**2*psi_til**2)/(4.0) #P(-)
      prob[4,:]= (psi_abs2[:,0] - 2.0*psi_til*post_state[0,:]*rpsi[:,0]+\
               2.0*psi_til*post_state[0,:]*ipsi[:,0]+\
               2.0*post_state[0,:]**2*psi_til**2)/(4.0) #P(L)
      prob[5,:]= (psi_abs2[:,0] - 2.0*psi_til*post_state[0,:]*rpsi[:,0]-\
               2.0*psi_til*post_state[0,:]*ipsi[:,0]+\
               2.0*post_state[0,:]**2*psi_til**2)/(4.0) #P(R)
      return np.real(prob)

def get_bisection(state,prob,dim):
   # to calculate bisection

   if typex(state)=='oper':
      rs = np.zeros((6,dim,dim))
      for i in range(6):
         for j in range(dim):
            for k in range(dim):
               rs[i,j,k] = randunit()/prob[i,j,k]
   else:
      rs = np.zeros((6,dim))
      for i in range(6):
         for j in range(dim):
            rs[i,j] = randunit()/prob[i,j]
   return rs

def get_rho(state,prob,niter,dim):
   # to calculate rho00 and rho01
   # just for mixed state

   N = np.int(niter/3) #we just meas. 2 bases {+,-} and {L,R}
   rho = np.zeros((2,dim,dim),dtype=complex)
   temp_rs = np.zeros((N,6,dim,dim))

   for i in range(N):
      temp_rs[i,:,:,:] = get_bisection(state,prob,dim)

   ave_rs = np.zeros((6,dim,dim))
   for i in range(6):
      for j in range(dim):
         for k in range(dim):
            ave_rs[i,j,k] = cdf(temp_rs[:,i,j,k])
   for x in range(dim):
      for p in range(dim):
         #rho01
         rho[0,x,p] = 1/2.0*(ave_rs[0,x,p]-ave_rs[1,x,p]-\
                           1j*(ave_rs[2,x,p]-ave_rs[3,x,p]))
         #rho11
         rho[1,x,p] = ave_rs[5,x,p]
   return rho

def get_qurec(post_state,theta,rho,dim):
   # to calculate the reconstructed state
   # just for mixed state
   restate = np.zeros((dim,dim),dtype=complex)
   for p in range(dim):
      for m in range(dim):
         for k in range(dim):
            restate[p,m] += np.exp(1j*2*pi*k*(p-m)/dim)*\
                           (rho[0,p,k]+rho[1,p,k])
         restate[p,m] *= 2.0/(dim*post_state[0,p]\
                            *post_state[0,m])
   #generate physical state
   restate = normx(restate)
   return restate

def get_pqurec(state,post_state,theta,niter,dim):
   """
   to get pure quantum state 
   Eqs. (16,17) in the main text
   >>>             (P+ - P- + 2P1)
       Re[psi]_n = ---------------   (16)
                  psi_til(1+delta_n) 

   >>>                PL - PR
       Im[psi]_n = ---------------   (17)
                  psi_til(1+deta_n)
   """
   niter = int(niter/3.)
   temp_rs = np.zeros((niter,6,dim))
   psi_til = abs(dot(post_state,state)[0,0])
   prob = get_prob(state,post_state,theta,dim)

   for i in range(niter):
      temp_rs[i,:,:] = get_bisection(state,prob,dim)
      
   ave_rs = np.zeros((6,dim))  
   for i in range(6):
      for j in range(dim):
         ave_rs[i,j] = cdf(temp_rs[:,i,j])
   
   rtemp = np.zeros(dim)
   itemp = np.zeros(dim)
   rnorm = 0.0
   for i in range(dim):
      rtemp[i] = (ave_rs[2,i]-ave_rs[3,i]+2*ave_rs[1,i])/\
                  (psi_til*post_state[0,i])
      itemp[i] = (ave_rs[4,i]-ave_rs[5,i])/\
                  (psi_til*post_state[0,i])
      rnorm += rtemp[i]**2+itemp[i]**2

   rtemp /= sqrt(rnorm)
   itemp /= sqrt(rnorm)
   
   restate = np.zeros((dim,dim),dtype=complex)
   for i in range(dim):
      for j in range(dim):
         restate[i,j] = (rtemp[i]+1j*itemp[i])*\
                     (rtemp[j]-1j*itemp[j])
   return restate



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                                x: quantum measurement, quantum metrology, 
                                quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""
# quantum toolbox for operators

__all__ = ['dotx', 'tensorx', 'ptracex','deposex', 'itensorx']

import numpy as np
from numpy import arccos
from cmath import phase
from numpy.linalg import multi_dot
from scipy import sparse

def dotx(*args):
    """
    calculate dot product of input operators
    Parametors
    ----------
    args : list or array

    Return
    ------
    dot product of input operators
    """
    if not args:
        raise TypeError("Please put at least one argument")

    return multi_dot(args)

def tensorx(*args):
    """
    calculate tensor product of input operators
    Parameters:
    ----------
    args: list or array
    
    Return:
    -------
    tensor prduct of input
    """
    if not args:
        raise TypeError("Please put at least one argument")

    if len(args) == 1 and \
        isinstance(args[0],(tuple,list,np.ndarray)):
        targs = args[0]
    else:
        targs = args
    
    t = targs[0]
    for i in range(1,len(targs)):
        t = np.kron(t,targs[i])
    return t

def ptracex(rho,sub_remove):
    """ Calculate the partial trace
    Parameters
    ----------
    rho: np.ndarray
        Density matrix
    *args: list
        Index of qubit to be kept after taking the trace
    Returns
    -------
    subrho: np.ndarray
        Density matrix after taking partial trace
    """
    n = int(np.log2(rho.shape[0]))
    a = 0
    for i in sub_remove:
        temp = _ptrace_(rho,n-a,i-a)
        #after each step, the number of qubits decreased by 1
        a += 1
    return temp
    
def _ptrace_(state,n,p):
    """
    ------------------------
    Input
    state: density matrix
    n: number of qubits
    p: traced qubit
    ------------------------
    Output: partially traced density matrix
    """
    d = 2**(n-1)
    temp = np.zeros((d,d),dtype=complex)
    for i in range(d):
        for j in range(d):
            temp[i,j] = state[_binary_(i,n,p,0),_binary_(j,n,p,0)]\
                    +state[_binary_(i,n,p,1),_binary_(j,n,p,1)]
    return temp

def _binary_(num,length,p,i):
    """
    First, represent number 'num' in binary string.
    Then, insert number 'i' into 'p' position of the string.
    ------------------------
    num: integer number
    length: number of qubits
    p: integer number
    i: either 0 or 1
    ------------------------
    Output: integer number
    """
    if (2**length < num):
        raise TypeError("Input can be truncated")
    if (length < p):
        print(length,p)
        raise TypeError("Out of range")
    if ((i != 0) and (i != 1)):
        raise TypeError("Inserted number must be 0 or 1")
    string = np.binary_repr(num,length)
    temp = string[:p]+str(i)+string[p:]
    result = int(temp,2)
    return result

def deposex(qubit):
    """
    depose a qubit into polar and azimuthal angles
    Parameters:
    ----------
    qubit
    
    Return:
    -------
    polar and azimuthal angles
    """
    polar = 2 * arccos(abs(qubit[0]))
    azimuthal = phase(qubit[1]) - phase(qubit[0])

    return float(polar), float(azimuthal)

def itensorx(oper,N):
    """calculate a list of tensors between oper and eye

    Args:
        oper (matrix): operator
        N (int): number of qubits
        
    Return
        list of tensors    
    """
    # convert oper to sparse matrix
    oper_csr = sparse.csr_matrix(oper)
    idn_csr = sparse.identity(2)
        
    ltensors = []
    for i in range(N):
        # set list of N opers with l[i] = oper
        itensors = [idn_csr] * N
        itensors[i] = oper_csr
        
        # compute tensor
        result = itensors[0]
        for j in range(1, N):
            result = sparse.kron(result, itensors[j])

        result_csr = sparse.csr_matrix(result.todense())  
        ltensors.append(result_csr)    
    return ltensors

def ftensorx(oper,N):
    """calculate full tensors of N opers
    i.e., oper \otimes oper \otimes ...

    Args:
        oper (matrix): operator
        N (int): number of qubits
        
    Return
        list of tensors    
    """
    # convert oper to sparse matrix
    oper_csr = sparse.csr_matrix(oper)
    
    ftensor = oper_csr
    for i in range(1,N):
        ftensor = sparse.kron(ftensor, oper_csr)  
    return ftensor


"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                                x: quantum measurement, quantum metrology, 
                                quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

__all__ = ['eyex','soper','sigmax','sigmay','sigmaz',
            'sigmap','sigmam','lowering','raising',
            'displacement','squeezing',
            'joper','jpauli','jnoper', 
            'dephasing_chl', 'bitflip_chl','phaseflip_chl', 
            'depolarizing_chl','markovian_chl',
            'nonmarkovian_chl']

import numpy as np
from numpy import conj
from scipy import sparse
import tqix 

def eyex(n):
    # to gererate an n x n identity matrix
    return np.identity(n)

#
# spin operater
#
def soper(s,*args):
    # spin matrices
    # represented in the Zeeman basis
    # s : float: 1/2, 1, 3/2,...
    # args : 'x','y','z','+','-'
    if not args:
        return soper(s,'x'),soper(s,'y'),soper(s,'z')
    if args[0] == '+':
        M = _sp(s)
    elif args[0] == '-':
        M = tqix.qobj.daggx(_sp(s))
    elif args[0] == 'x':
        M = 0.5 * (_sp(s) + tqix.qobj.daggx(_sp(s)))
    elif args[0] == 'y':
        M = -0.5 * 1j * (_sp(s) - tqix.qobj.daggx(_sp(s)))
    elif args[0] == 'z':
        M = _sz(s)
    else:
        raise TypeError('Invalid type')
    return M

def _sz(s):
    # define sz
    d = int(2*s + 1)
    data = np.zeros((d,d),dtype = complex)
    for i in range(0,d):
        for j in range(0,d):
            data[i,j] = tqix.utility.krondel(s-i,s-j)*(s-i)
    return data

def _sp(s):
    # define s+
    d = int(2*s + 1)
    data = np.zeros((d,d),dtype = complex)
    for i in range(0,d):
        for j in range(0,d):
            data[i,j] = tqix.utility.krondel(s-i,s-j+1)*\
            np.sqrt(s*(s+1)-(s-i)*(s-j))
    return data

#
# Pauli matrices
#
def sigmax():
    # Pauli matrix sigma x
    return 2.0 * soper(1.0 / 2, 'x')

def sigmay():
    # Pauli matrix sigma y
    return 2.0 * soper(1.0 / 2, 'y')

def sigmaz():
    # Pauli matrix sigma z
    return 2.0 * soper(1.0 / 2, 'z')

def sigmap():
    # Sigma plus
    return (sigmax() + 1j*sigmay())/2.0

def sigmam():
    # Sigma minus
    return (sigmax() -1j*sigmay())/2.0

#
# annihilation operator or the lowering operator
#
def lowering(d):
    # lowering operator
    # d : dimension
    if not isinstance(d,(int, np.integer)):
        raise ValueError("dimension must be integer")
    M = np.zeros((d,d),dtype = float)
    for i in range (d-1):
        M[i,i+1] = np.sqrt(i + 1.0)
    return M

#
# creation operator or the raising operator
#
def raising(d):
    # raising operator
    # d : dimension
    if not isinstance(d,(int, np.integer)):
        raise ValueError("dimension must be integer")
    return tqix.qobj.daggx(lowering(d))
    
#
# displacement operator
#
def displacement(d,alpha):
    power = alpha*raising(d) - conj(alpha)*lowering(d)
    M = tqix.qobj.expx(power)
    return M

#
# squeezing operator
#
def squeezing(d,beta):
    power = 0.5 * (conj(beta)*tqix.qtool.dotx(lowering(d),lowering(d))\
            - beta*tqix.qtool.dotx(raising(d),raising(d)))
    M = tqix.qobj.expx(power)
    return M
#
# collective spin operators
#
def joper(N,*args):
    # collective spin operators
    # N: integer: number of spin-1/2
    # args: 'x','y','z','+','-'

    if not args:
        return joper(N,'x'),joper(N,'y'),joper(N,'z')
    if args[0] == 'x':
        s = 0.0
        for i in range(N): 
            s += jnoper(N,i,'x')           
        return s/2. 

    if args[0] == 'y':
        s = 0.0
        for i in range(N):
            s += jnoper(N,i,'y')
        return s/2.

    if args[0] == 'z':
        s = 0.0
        for i in range(N):
            s += jnoper(N,i,'z')
        return s/2. 

    if args[0] == 'p':
        return joper(N,'x')+1j*joper(N,'y')

    if args[0] == 'm':
        return joper(N,'x')-1j*joper(N,'y')

def jpauli(N,*args):
    # collective Pauli operators
    # N: integer: number of spin-1/2
    # args: 'x','y','z','+','-'

    if not args:
        return jpauli(N,'x'),jpauli(N,'y'),jpauli(N,'z')
    if args[0] == 'x':
        s = 0.0
        for i in range(N): 
            s += jnoper(N,i,'x')           
        return s 

    if args[0] == 'y':
        s = 0.0
        for i in range(N):
            s += jnoper(N,i,'y')
        return s

    if args[0] == 'z':
        s = 0.0
        for i in range(N):
            s += jnoper(N,i,'z')
        return s

    if args[0] == 'p':
        return jpauli(N,'x')+1j*jpauli(N,'y')

    if args[0] == 'm':
        return jpauli(N,'x')-1j*jpauli(N,'y')

def jnoper(N,i,*args):
    # spin jn operators @ site i
    # N: interager: number of spin-1/2
    # args: 'x','y','z','+','-'

    if not args:
        raise TypeError("please put 'x' or 'y' or 'z'")
    if args[0] == 'x':
        if i < 0 or i >= N:
            raise TypeError('i out of range: [0,N-1]')
        elif i == 0:
            res = sigmax()
            for j in range(1,N):
                res = tqix.qtool.tensorx(res,eyex(2))
        else:
            res = eyex(2)
            for j in range(1,i):
                res = tqix.qtool.tensorx(res,eyex(2))
            res = tqix.qtool.tensorx(res,sigmax())
            for j in range(i,N-1):
                res = tqix.qtool.tensorx(res,eyex(2))
        return res

    if args[0] == 'y':
        if i < 0 or i >= N:
            raise TypeError('i out of range: [0,N-1]')
        elif i == 0:
            res = sigmay()
            for j in range(1,N):
                res = tqix.qtool.tensorx(res,eyex(2))
        else:
            res = eyex(2)
            for j in range(1,i):
                res = tqix.qtool.tensorx(res,eyex(2))
            res = tqix.qtool.tensorx(res,sigmay())
            for j in range(i,N-1):
                res = tqix.qtool.tensorx(res,eyex(2))
        return res

    if args[0] == 'z':
        if i < 0 or i >= N:
            raise TypeError('i out of range: [0,N-1]')
        elif i == 0:
            res = sigmaz()
            for j in range(1,N):
                res = tqix.qtool.tensorx(res,eyex(2))
        else:
            res = eyex(2)
            for j in range(1,i):
                res = tqix.qtool.tensorx(res,eyex(2))
            res = tqix.qtool.tensorx(res,sigmaz())
            for j in range(i,N-1):
                res = tqix.qtool.tensorx(res,eyex(2))
        return res

# add noise channel via kraus operators
def dephasing_chl(x, lamb):
    """Add dephasing to the system

    Args:
        - x quantum system state
        - lamb: noise
    
    Return
        - x
    """ 
    # number of qubits 
    N = tqix.qobj.qubitx(x)
    
    # kraus operators
    kraus1 = np.array([[1, 0],[0, np.sqrt(1 - lamb)]])
    kraus2 = np.array([[0, 0],[0, np.sqrt(lamb)]])
    
    fkraus1 = tqix.qtool.ftensorx(kraus1, N)
    fkraus2 = tqix.qtool.ftensorx(kraus2, N)
    
    # apply Kraus to qobj x
    x_csr = sparse.csr_matrix(tqix.operx(x))
    x_csr = fkraus1 @ x_csr @ fkraus1 + fkraus2 @ x_csr @ fkraus2
    
    return x_csr

def bitflip_chl(x, lamb):
    """Add bitflip to the system

    Args:
        - x quantum system state
        - lamb: noise
    
    Return
        - x (sparse matrix)
    """ 
    # number of qubits 
    N = tqix.qobj.qubitx(x)
    
    # kraus operators
    kraus1 = np.sqrt(1-lamb) * np.array([[1, 0], [0, 1]])
    kraus2 = np.sqrt(lamb) * np.array([[0, 1], [1, 0]])
    
    fkraus1 = tqix.qtool.ftensorx(kraus1, N)
    fkraus2 = tqix.qtool.ftensorx(kraus2, N)
    
    # apply Kraus to qobj x
    x_csr = sparse.csr_matrix(tqix.operx(x))
    x_csr = fkraus1 @ x_csr @ fkraus1 + fkraus2 @ x_csr @ fkraus2
    
    return x_csr

def phaseflip_chl(x, lamb):
    """Add phaseflip to the system

    Args:
        - x quantum system state
        - lamb: noise
    
    Return
        - x (sparse matrix)
    """ 
    # number of qubits 
    N = tqix.qobj.qubitx(x)
    
    # kraus operators
    kraus1 = np.sqrt(1-lamb) * np.array([[1, 0], [0, 1]])
    kraus2 = np.sqrt(lamb) * np.array([[1, 0], [0, -1]])
    
    fkraus1 = tqix.qtool.ftensorx(kraus1, N)
    fkraus2 = tqix.qtool.ftensorx(kraus2, N)
    
    # apply Kraus to qobj x
    x_csr = sparse.csr_matrix(tqix.operx(x))
    x_csr = fkraus1 @ x_csr @ fkraus1 + fkraus2 @ x_csr @ fkraus2
    
    return x_csr

def depolarizing_chl(x, lamb):
    """Add depolarizing to the system

    Args:
        - x quantum system state
        - lamb: noise
    
    Return
        - x (sparse matrix)
    """ 
    # number of qubits 
    N = tqix.qobj.qubitx(x)
    
    # kraus operators
    kraus1 = np.sqrt(1-lamb) * np.array([[1, 0], [0, 1]])
    kraus2 = np.sqrt(lamb/3.) * np.array([[0, 1], [1, 0]])
    kraus3 = np.sqrt(lamb/3.) * np.array([[0, -1j], [1j, 0]])
    kraus4 = np.sqrt(lamb/3.) * np.array([[1, 0], [0, -1]])
    
    fkraus1 = tqix.qtool.ftensorx(kraus1, N)
    fkraus2 = tqix.qtool.ftensorx(kraus2, N)
    fkraus3 = tqix.qtool.ftensorx(kraus3, N)
    fkraus4 = tqix.qtool.ftensorx(kraus4, N)
    
    # apply Kraus to qobj x
    x_csr = sparse.csr_matrix(tqix.operx(x))
    x_csr = fkraus1 @ x_csr @ tqix.qobj.daggx(fkraus1) \
            + fkraus2 @ x_csr @ tqix.qobj.daggx(fkraus2) \
            + fkraus3 @ x_csr @ tqix.qobj.daggx(fkraus3) \
            + fkraus4 @ x_csr @ tqix.qobj.daggx(fkraus4)
    
    return x_csr

def markovian_chl(x, t, y):
    """Add markovian to the system

    Args:
        - x quantum system state
        - t: time
        - y: noise
    
    Return
        - x (sparse matrix)
    """ 
    # number of qubits 
    N = tqix.qobj.qubitx(x)
    
    # kraus operators
    qt = 1 - np.exp(-y*t)
    kraus1 = np.array([[np.sqrt(1-qt) ,0], [0,1]])
    kraus2 = np.array([[np.sqrt(qt),0], [0,0]])
    
    fkraus1 = tqix.qtool.ftensorx(kraus1, N)
    fkraus2 = tqix.qtool.ftensorx(kraus2, N)
    
    # apply Kraus to qobj x
    x_csr = sparse.csr_matrix(tqix.operx(x))
    x_csr = fkraus1 @ x_csr @ fkraus1 + fkraus2 @ x_csr @ fkraus2
    
    return x_csr

def nonmarkovian_chl(x, t, y, tc = 20.0):
    """Add nonmarkovian to the system

    Args:
        - x quantum system state
        - t: time
        - y: noise
    
    Return
        - x (sparse matrix)
    """ 
    # number of qubits 
    N = tqix.qobj.qubitx(x)
    
    # kraus operators
    # tc = 20.0 #fixed in rXiv:2305.08289
    qt = 1 - np.exp(-y*t*t/(2*tc))
    
    kraus1 = np.array([[np.sqrt(1-qt) ,0], [0,1]])
    kraus2 = np.array([[np.sqrt(qt),0], [0,0]])
    
    fkraus1 = tqix.qtool.ftensorx(kraus1, N)
    fkraus2 = tqix.qtool.ftensorx(kraus2, N)
    
    # apply Kraus to qobj x
    x_csr = sparse.csr_matrix(tqix.operx(x))
    x_csr = fkraus1 @ x_csr @ fkraus1 + fkraus2 @ x_csr @ fkraus2
        
    return x_csr



# -*- coding: utf-8 -*-
""" Toolbox for quantum metrology
    Calculate Holevo bound using semidefinite programing
    Calculate SLD and RLD bounds
    Author: Le Bin Ho @ 2022
"""

import qiskit
import tqix.vqa.circuits
import tqix.vqa.constants
import tqix.vqa.vqm

import numpy as np
from numpy.linalg import inv, multi_dot, norm
from scipy.linalg import sqrtm, solve_sylvester
import copy

from tqix.qobj import eigenx,dotx,daggx

def sld_qfim(qc, qcirs, method = 'eigens'):
    
    """calculate QFIM using SLD with different methods
    
    Args:
        - qc : initial circuit
        - qcirs: set of circuits (we always qc_add as model, no call here)

    Returns:
        - sld_qfim
    """
    if method == 'eigens':
        return sld_qfim_eigens(qc, qcirs)
    elif method == 'inverse':
        return sld_qfim_inverse(qc, qcirs)
    elif method == 'sylvester':
        return sld_qfim_sylvester(qc, qcirs)
    else:
        raise ValueError(
            "Method is not avilable: please use 'eigens', 'inverse', or 'sylvester' ")
    
    
    
def sld_qfim_inverse(qc, qcirs):
    
    """ calculate the QFIM bases SLD 
    
    Args:
        - qc : initial circuit
        - qcirs: set of circuits (we always qc_add as model, no call here)

    Returns:
        - qfim 
    """ 
    
    cir = tqix.vqa.vqm.qc_add(qc.copy(), qcirs)
    rho = tqix.vqa.circuits.state_density(cir.copy()) #rho after evolve
    grho = _grad_rho(qc.copy(), qcirs)
    
    d = len(grho) # number of paramaters
    H = np.zeros((d,d), dtype = complex)
    invM = _inv_M(rho)
    
    vec_grho = []  
    for i in range(d):
        vec_grho.append(_vectorize(grho[i]))

    for i in range(d):
        for j in range(d): 
            H[i,j] = 2*multi_dot([np.conjugate(vec_grho[i]).T, invM, vec_grho[j]]) 
    return np.real(H)  


def sld_qfim_sylvester(qc, qcirs):
    
    """ calculate the QFIM bases SLD 
    
    Args:
        - qc : initial circuit
        - qcirs: set of circuits (we always qc_add as model, no call here)

    Returns:
        - qfim 
    """ 
    
    cir = tqix.vqa.vqm.qc_add(qc.copy(), qcirs)
    rho = tqix.vqa.circuits.state_density(cir.copy()) #rho after evolve
    grho = _grad_rho(qc.copy(), qcirs)
    
    d = len(grho) # number of paramaters
    H = np.zeros((d,d), dtype = complex)      
    L = sylvester(rho,rho,grho)
        
    for i in range(d):
        for j in range(d):
            H[i,j] = 0.5*np.trace(rho @ AntiCommutator(L[i],L[j]))

    return np.real(H)     


def sld_qfim_eigens(qc, qcirs):
    
    """ calculate the QFIM bases eigenvalues
    
    Args:
        - qc : initial circuit
        - qcirs: set of circuits (we always qc_add as model, no call here)

    Returns:
        - qfim 
    """ 
    
    cir = tqix.vqa.vqm.qc_add(qc.copy(), qcirs)
    rho = tqix.vqa.circuits.state_density(cir.copy()) #rho after evolve
    grho = _grad_rho(qc.copy(), qcirs)
    
    d = len(grho) # number of paramaters
    H = np.zeros((d,d), dtype = complex)      
    
    eigens = eigenx(rho)
        
    for k in range(d):
        for l in range(d):
            H[k,l] = 0.0
            for i in range(len(eigens[0])):
                for j in range(len(eigens[0])):
                    de = eigens[0][i] + eigens[0][j]
                    if de > 10e-15:
                        num1 = dotx(daggx(eigens[1][i]),grho[k],eigens[1][j])
                        num2 = dotx(daggx(eigens[1][j]),grho[l],eigens[1][i])
                        H[k,l] += 2*np.real(num1*num2)/de
                        
    return np.real(H)  


def sld_bound(qc, qcirs):
    """ return the SLD bound 
    
    Args:
        - qc
        - qcirs

    Returns:
        - sld bound
    """
    
    d = len(qcirs[1][2]) #[1]:u_phase,[1][2]:phases
    W = np.identity(d)
    
    sld = sld_qfim(qc.copy(), qcirs)
    return np.real(np.trace(W @ inv(sld + np.eye(len(sld)) * 10e-10)))


def rld_qfim(qc, qcirs): 
    
    """ calculate the QFIM bases RLD 
    
    Analytical:
        H_ij = tr(rho*L_i*L_j^\dag)
            = tr(L_j^\dag*rho*L_i)

        submit rho*L_i = der_i(rho) : this is rld
        we get: H_ij = tr(L_j^\dag*dev_i(rho))
        
        apply tr(A^dag*B) = vec(A)^dag*vec(B), 
        we have: H_ij = vec(L_j)^dag*vec[dev_i(rho)]
        here: vec[L_j)] = (I x rho)^{-1)*vec(L_j)
        
    Args:
        - qc
        - qcirs

    Returns:
        - qfim via rld   
    """ 
    qc_func = tqix.vqa.vqm.qc_add(qc.copy(), qcirs)
    rho = tqix.vqa.circuits.state_density(qc_func.copy())
    grho = _grad_rho(qc.copy(), qcirs)
    
    d = len(grho) # number of estimated parameters  
    R = np.zeros((d,d), dtype = complex)
    IR = _i_R(rho)
    
    vec_grho = []
    for i in range(d):
        vec_grho.append(_vectorize(grho[i]))
    
    for i in range(d):
        for j in range(d): 
            vecLj = np.conjugate(multi_dot([IR, vec_grho[j]])).T
            R[i,j] = multi_dot([vecLj, vec_grho[i]])  
            
    return R


def rld_bound(qc, qcirs):
    """ return the SLD bound 
    
    Args:
        - qc
        - cirs
        - params
        - coefs
        - W: weight matrix

    Returns:
        - rld bound
    """
    
    d = len(qcirs[1][2]) #[1]:u_phase,[1][2]:phases
    W = np.identity(d)
        
    rld = rld_qfim(qc, qcirs)
    invrld = inv(rld + np.eye(len(rld)) * 10e-10)
    R1 = np.trace(W @ np.real(invrld))
    R2 = norm(multi_dot([sqrtm(W), np.imag(invrld),sqrtm(W)])) 
    
    return R1 + R2


def cfim(qc, qcirs):
    """ return the classical fisher information matrix
    
    Args:
        - qc
        - qcirs
        
    Returns:
        - cfim
    """
    # measurements
    qc_copy = tqix.vqa.vqm.qc_add(qc.copy(), qcirs)
    pro = tqix.vqa.circuits.measure_born(qc_copy)
    
    dpro = []
    d = len(qcirs[1][2]) #[1]:u_phase,[1][2]:phases
    s = tqix.vqa.constants.step_size
    
    #remove zeros indexs
    idxs = np.where(pro<=10e-18)
    pro = np.delete(pro, idxs)

    
    for i in range(0, d):
        # We use the parameter-shift rule explicitly
        # to compute the derivatives
        qcirs1, qcirs2 = copy.deepcopy(qcirs), copy.deepcopy(qcirs)    
        qcirs1[1][2][i] += s #[1]:u_phase,[1][2]:phases
        qcirs2[1][2][i] -= s
        
        plus = tqix.vqa.vqm.qc_add(qc.copy(), qcirs1) 
        minus = tqix.vqa.vqm.qc_add(qc.copy(), qcirs2)
        gr = (tqix.vqa.circuits.measure_born(plus)-tqix.vqa.circuits.measure_born(minus))/(2*s)
        gr = np.delete(gr, idxs)
        dpro.append(np.array(gr))
    
    matrix = np.zeros((d,d), dtype = float)
    for i in range(d):
        for j in range(d):      
            matrix[i,j] = np.sum(dpro[i] * dpro[j] / pro)

    return matrix


def cls_bound(qc, qcirs):
    """ return the classical bound 
    
    Args:
        - qc
        - qcirs
        - W: weight matrix

    Returns:
        - rld bound
    """
    
    #list2str = list(map(lambda f: f.__name__, cirs)) 
    #idx = list2str.index('u_phase') 
    #d = len(params[idx])
        
    clf = cfim(qc.copy(), qcirs)
    W = np.identity(len(clf))
    
    return np.trace(W @ inv(clf + np.eye(len(clf)) * 10e-10))


def _vectorize(rho):
    # return a vectorized of rho
    # rho: a matrices (data)
    vec_rho = np.reshape(rho, (len(rho)**2,1), order='F')
    return vec_rho


def _grad_rho(qc, qcirs):
    """ calculate drho by parameter-shift rule
    
    Args:
        - qc: initial cuircuit (we need to reset for rho(+) and rho(-)
        - qcirs: circuits
        
    Return:
        - gradient of state density w.r.t. all phases
    """
    
    dp = [] #array of matrices rho
    s = tqix.vqa.constants.step_size 
    
    for i in range(0, len(qcirs[1][2])):
        qcirs1, qcirs2 = copy.deepcopy(qcirs), copy.deepcopy(qcirs)       
        qcirs1[1][2][i] += s
        qcirs2[1][2][i] -= s
        
        plus = tqix.vqa.vqm.qc_add(qc.copy(), qcirs1) 
        minus = tqix.vqa.vqm.qc_add(qc.copy(), qcirs2)       
        dp.append((tqix.vqa.circuits.state_density(plus)-tqix.vqa.circuits.state_density(minus))/(2*s))
        
    return dp


def _inv_M(rho, epsilon = 10e-10): 
    """ return inverse matrix M 
        M = rho.conj()*I + I*rho.conj()
    
    Args:
        - quantum state rho (data)

    Returns:
        - inverse matrix M 
    """
    
    d = len(rho)
    M = np.kron(np.conj(rho), np.identity(d)) + np.kron(np.identity(d), rho)
    return inv(M + np.eye(len(M)) * epsilon)


def _i_R(rho, epsilon = 10e-10):
    """ return inverse of matrix R 
        R = I*rho.conj()
    
    Args:
        - quantum state rho (data)

    Returns:
        - inverse matrix R 
    """    
    d = len(rho)
    R = np.kron(np.identity(d), rho)
    return inv(R) #inv(R + np.eye(len(R)) * epsilon)


def sylvester(A,B,C):
    """ solve the sylvester function:
        AX + XB = 2*C # the symmetric logarithmic derivative
        here A,B = rho for SLD, i.e., rho @ L + L @ rho = 2*drho
            A = rho, B = 0 for RLD, i.e., rho @ L = 2*drho
            C = drho
    Args:
        - rho: input quantum state
        - drho: input derivative of quantum state
    Returns:
        - X operator
    """
    
    lenC = len(C) #we have several drho (gradient of rho)
    X = []
    
    for i in range(lenC):
        L = solve_sylvester(A, B, 2.0*C[i])
        X.append(L)
        
    #print(X)  
    return X


def AntiCommutator(A,B):
    # AB + BA
    return A @ B + B @ A                     


def calculate_SLD(qc, qcirs):
    # to calculate SLD
        
    cir = tqix.vqa.vqm.qc_add(qc.copy(), qcirs)
    rho = tqix.vqa.circuits.state_density(cir.copy()) #rho after evolve
    grho = _grad_rho(qc.copy(), qcirs)
    
    d = len(grho) # number of paramaters     
    Ls = []
    eigens = eigenx(rho)
    
    for i in range(d):
        L = np.zeros((len(rho),len(rho)), dtype = complex) 
        for k in range(len(eigens[0])):
            for l in range(len(eigens[0])):
                de = eigens[0][k] + eigens[0][l]
                if de > 10e-15:
                    coe1 = dotx(daggx(eigens[1][k]),grho[i],eigens[1][l])/de
                    coe2 = eigens[1][k]@daggx(eigens[1][l])
                    L += 2*coe1*coe2
        Ls.append(L)                
    return Ls


from tqix import *
import numpy as np

psi = zbasis(3,2)

# 3d visualization
x = [-3, 3]
y = [-3, 3]
husimi_3d(psi, x ,y ,cmap = cmindex(1),fname='husimi3d.eps')
wigner_3d(psi, x ,y ,cmap = cmindex(1),fname='wigner3d.eps')

# Bloch sphere visualization
THETA = [0, np.pi]
PHI = [0, 2* np.pi]
husimi_spin_3d(psi, THETA ,PHI ,cmap =cmindex(1),fname = 'husimi_sphere.eps')
wigner_spin_3d(psi, THETA ,PHI ,cmap =cmindex(1),fname = 'wigner_sphere.eps')




from tqix.pis import *
from tqix import *
import numpy as np
from matplotlib import pyplot as plt
import torch 
import numpy as np 

def cost_function(theta,use_gpu=False):
    qc = circuit(N,use_gpu=use_gpu)
    qc.RN(np.pi/2,0)
    qc.OAT(theta[0],"Z") 
    qc.TNT(theta[1],omega=theta[1],gate_type="ZX")
    qc.TAT(theta[2],"ZY")
    if use_gpu:
        loss = torch.real(get_xi_2_S(qc))
    else:
        loss = np.real(get_xi_2_S(qc))
    return  loss

N=100 #number of qubits
route = (("RN2",),("OAT","Z"),("TNT","ZX"),("TAT","ZY")) #define layers for QNG algorithm
loss_dict = {}

#function to optimize circuit of sparse array
def sparse(optimizer,loss_dict,mode):
    objective_function = lambda params: cost_function(params) 
    init_params = [0.00195902, 0.14166777, 0.01656466] #random init parameters for circuit
    _,_,_,loss_hist,time_iters = fit(objective_function,optimizer,init_params,return_loss_hist=True,return_time_iters = True)
    loss_dict[mode] = loss_hist
    return loss_dict,time_iters

#function to optimize circuit of tensor
def tensor(optimizer,loss_dict,mode):
    objective_function = lambda params: cost_function(params,use_gpu=True) 
    init_params = [0.00195902, 0.14166777, 0.01656466] #random init parameters for circuit
    init_params = torch.tensor(init_params).to('cuda').requires_grad_()
    _, _,_,loss_hist,time_iters = fit(objective_function,optimizer,init_params,return_loss_hist=True,return_time_iters = True)
    loss_dict[mode] = loss_hist
    return loss_dict,time_iters

#QNG
optimizer = GD(lr=0.03,eps=1e-10,maxiter=200,use_qng=True,route=route,tol=1e-19,N=N)
loss_dict,_ = tensor(optimizer,loss_dict,"tensor_qng")
print(loss_dict)
#GD
optimizer = GD(lr=0.0001,eps=1e-10,maxiter=200,tol=1e-19,N=N)
loss_dict,_ = tensor(optimizer,loss_dict,"tensor_gd")

#ADAM 
optimizer = ADAM(lr=0.01,eps=1e-10,amsgrad=False,maxiter=200)
loss_dict,sparse_times = sparse(optimizer,loss_dict,"sparse_adam")

optimizer = ADAM(lr=0.001,eps=1e-10,amsgrad=False,maxiter=200)
loss_dict,tensor_times = tensor(optimizer,loss_dict,"tensor_adam")

#plot loss values with respect to iterations
ax = plt.gca() 
ax.plot(range(len(loss_dict["tensor_qng"] )),loss_dict["tensor_qng"] ,'c-*',label=r'$QNG;\eta=0.1$')
ax.plot(range(len(loss_dict["tensor_gd"])),loss_dict["tensor_gd"],'y-^',label=r'$GD;\eta=0.03$')
ax.plot(range(len(loss_dict["tensor_adam"])),loss_dict["tensor_adam"],'g-o',label=r'$Adam;\eta=0.01$')


ax.set_xlabel("iterations")
ax.set_ylabel(r"$C(\theta)$")
ax.set_xlim(0,130)
ax.set_ylim(0,15)
lgd = ax.legend(loc='center left', bbox_to_anchor=(1, 0.25))
plt.savefig("./loss_bm.eps", bbox_extra_artists=(lgd,), bbox_inches='tight')

#for compare running tume between using tensor and sparse array structure, we plot ADAM as an example
cumsum_vqa_time_res_sparse = np.cumsum(loss_dict['sparse_adam'])
cumsum_vqa_time_res_tensor = np.cumsum(loss_dict['tensor_adam'])
from matplotlib import pyplot as plt
ax = plt.gca() 
ax.plot(range(len(cumsum_vqa_time_res_sparse)),cumsum_vqa_time_res_sparse,'g--',label=r'$CPU$')
ax.plot(range(len(cumsum_vqa_time_res_tensor)),cumsum_vqa_time_res_tensor,'r-',label=r'$GPU$')

ax.set_xlabel("iterations")
ax.set_ylabel("time(s)")
ax.set_ylim(0,100)
lgd = ax.legend(loc='upper right')
plt.savefig("./timetensorsparse.eps", bbox_extra_artists=(lgd,), bbox_inches='tight')



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                                x: quantum measurement, quantum metrology, 
                                quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors:
>>> all rights reserved
________________________________
"""

#from numpy import
import numpy as np
from tqix.qobj import *
from tqix.pis.util import *
from scipy.sparse import csc_matrix
from tqix.pis import *
import torch 
__all__ =['circuit','sobj',
            'dbx','dicke_ghz']

def circuit(N,**kwargs):
    """

    :param N: number of qubits
    :type N: int
    :return: spin state of circuit 
    :rtype: instance of sobj
    """    
    use_gpu = kwargs.pop('use_gpu', False)
    init_state = kwargs.pop('initial_state', None)
    num_process = kwargs.pop('num_process', None)
    if use_gpu:
        if torch.cuda.is_available():
            device = "cuda"
        else:
            print("your machine does not utilize gpu, the program will use torch tensor type on cpu")
            device = "cpu"
    else:
        device = None

    if not init_state:
        j = N/2
        psi = dbx(j,-j) # all spins down
        if use_gpu:
            psi = psi.todense()
            return sobj(torch.tensor(operx(psi)).to(device),N,use_gpu=use_gpu,device=device,num_process=num_process)
        return sobj(operx(psi).tolist(),N,use_gpu=use_gpu,device=device,num_process=num_process) 
    else:
        return sobj(init_state,N,use_gpu=use_gpu,device=device,num_process=num_process)

class sobj(Gates):
    """
    class stores state of qubits
    """    
    # to crate a spin-object
    def __init__(self,state,N,use_gpu=None,device=None,num_process=None):
        """

        :param state: state of qubits
        :type state: ndarray,tensor,sparse
        :param N: number of qubits
        :type N: int
        :param use_gpu: if use gpu, defaults to None
        :type use_gpu: bool, optional
        :param device: name of compute device, defaults to None
        :type device: str, optional
        :param num_process: number of processes, defaults to None
        :type num_process: int, optional
        """        
        super().__init__()
        self.state = state
        self.N = N
        self.use_gpu = use_gpu
        self.device = device
        self.num_process = num_process 
        
    def print_state(self):
        state = self.state
        return state

def dbx(j,m):
    """
    create dicke basis with pure state

    input: j,m 
    
    output: a vector basis

    :param j: index j
    :type j: int
    :param m: index m
    :type m: int
    :raises ValueError: j must in bound -j ≤ m ≤ j
    :return: state in dicked basis
    :rtype: ndarray, tensor, sparse
    """    
    # creat dicke basis with pure state
    # input: j,m 
    # output: a vector basis

    if m > j or m < -j:
        raise ValueError('j must in bound -j ≤ m ≤ j')
    
    dim = int(2*j + 1)
    state = np.zeros((dim,1))
    offset = get_vidx(j,m) #get vector's index
    state[offset,0] = 1.0
    return csc_matrix(state)

def dicke_ghz(N):
    """
    :param N: number of qubits
    :type N: int
    :return: state in dicked ghz basis
    :rtype: ndarray, tensor, sparse
    """    
    # this is an example to get ghz state
    # def: GHZ = (|N/2,N/2> + |N/2,-N/2>)√2
        
    j = N/2
    m = j
    m1 = -j
    
    e1 = dicke_bx(N,{(j,m,m):1})
    e2 = dicke_bx(N,{(j,m,m1):1})
    e3 = dicke_bx(N,{(j,m1,m):1})
    e4 = dicke_bx(N,{(j,m1,m1):1})
    
    return csc_matrix(0.5*(e1+e2+e3+e4))



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

# description: math tools for a quantum object x

__all__ = ['qx','typex','shapex','qubitx',
            'isqx','isbrax','isketx','isoperx','ishermx','isnormx',
            'operx','diagx','conjx','transx','daggx','tracex','eigenx',
            'groundx','expx','sqrtx','l2normx','normx']

from numpy import (allclose,array_equal)
import numpy as np
from tqix.qtool import dotx
import scipy
import torch 

def qx(x):
      """ 
      to turn x into a quantum object.
      we define a quantum object is an array dims [[n],[m]]
      or shape (n, m)
      >>> for example: 
         qx([[1],[2],[3]]) will return a ket state:
         => [[1]
            [2]
            [3]]
      """
      if isinstance(x,(int,float,complex)):
         return np.array([[x]])
      elif isinstance(x,(tuple,list)):
         if len(get_shape(x)) == 1: #1d list
            qox = np.array([x])
         else:
            qox = np.array(x)
         return qox 
      else: # ndarray already 
         qox = np.array([x]) if x.ndim == 1 else np.array(x)
         return qox

def typex(x):
      """
      return a quantum type of x:
      bar, ket, oper
      """
      if scipy.sparse.issparse(x):
         x = x.toarray()
      if isqx(x):
         if x.shape[0]!=1:
            return 'oper' if x.shape[1] != 1 else 'ket'
         elif x.shape == (1,1):
            return 'oper'
         else:
            return 'bra'
      else:
         msg = 'not a quantum object.'
         raise TypeError(msg)

def shapex(x):
   """
   to return the dimension of x
   """
   if torch.is_tensor(x):
      x = x.detach().cpu().numpy()
   if scipy.sparse.issparse(x):
      x = x.toarray()
   if isqx(x):
      return x.shape
   else:
      msg = 'not a quantum object.'
      raise TypeError(msg)

def qubitx(x):
   """ convert shapex(x) to number of qubits if applicable

   Args:
      x (oper): quantum object
      
   Return: number of qubits    
   """
   
   dim = shapex(x)
   N = np.log2(dim[0])
   if N.is_integer():
      return int(N)
   else:
      msg = 'the quantum object is not formed from qubits'
      raise TypeError(msg)

def isqx(x):
   """
   to check if x is a quantum object or not
   """
   if isinstance(x,np.ndarray):
      return True if x.ndim == 2 else False
   else:
      return False

def isbrax(x):
   """
   to check if quantum object x is bra vector
   """
   return True if typex(x)=='bra' else False

def isketx(x):
   """
   to check if quantum object x is ket vector
   """
   return True if typex(x)=='ket' else False

def isoperx(x):
   """
   to check if x is operator (mixed state, Hamiltonian,...)
   """
   return True if typex(x)=='oper' else False

def ishermx(x):
   """
   to check if x is Hermit or not (for oper only)
   """
   if typex(x) == 'oper':
      return True if array_equal(daggx(x),x) else False
   else:
      msg = 'not an operator'
      raise TypeError(msg)

def isnormx(x):
   """ 
   to check if quantum object x is normalzed or not
   """
   if typex(x) != 'oper':
      opx = operx(x)
   else:
      opx = x
   if opx.shape[0] == opx.shape[1]:
      return True if allclose(tracex(opx),1.0) else False
   else:
      msg = 'not a square matrix'
      raise TypeError(msg)

def operx(x):
   """
   to convert a bra or ket vector into oper
   """
   if typex(x) == 'oper':
      return qx(x)
   elif typex(x) == 'bra':
      return qx(dotx(daggx(x),x))
   else:
      return qx(dotx(x,daggx(x)))

def diagx(x):
   # return diagonal of x
   w,v = eigenx(x)
   return np.diag(w)

def conjx(x):
   return np.conj(x)

def transx(x):
   return np.transpose(x)

def daggx(x):
   if torch.is_tensor(x):
      return torch.transpose(torch.conj(x),0,1)
   x = conjx(x)
   return transx(x)

def tracex(x):
   """
   to calculate trace for x
   """
   if typex(x) == 'oper':
      return np.trace(x)
   else:
      msg = 'not a square matrix.'
      raise TypeError(msg)

from numpy import linalg as LA
def eigenx(x):
   """
   eigenvalue and eigenvector

   Parameters:
   -----------
   x : operator

   Returns:
   --------
   w : [..., ...] ndarray
   vk : [[...,...],[...,...]] matrix
      vk[i] is the normalized "ket" eigenvector
      accroding to eigenvelue w[i]
   """
   if typex(x) != 'oper':
      raise TypeError('not an oper')
   else:
      if x.shape[0] != x.shape[1]:
         raise TypeError('not a square oper')
      else:
         w, v = LA.eigh(x)
         vk = []
         for i in range (len(w)):
            vk.append(transx(qx(v[:,i])))
         return w, vk

def groundx(x):
   """
   get ground state for a given Hamiltonian

   Parameters:
   -----------
   x : operator

   Returns:
   --------
   the normalized "ket" eigenvector
   accroding to minimum eigenvelue
   """

   w, vk = eigenx(x)
   m = np.argmin(w)
   return vk[m]

from scipy.linalg import expm,sqrtm
def expx(x):
   # exponential x
   if typex(x) != 'oper':
      raise TypeError('not an oper')
   else:
      return expm(x)

def sqrtx(x):
   # square root x
   if typex(x) != 'oper':
      raise TypeError('not an oper')
   else:
      return sqrtm(x)

def l2normx(x):
   """
   to calculate norm 2 of a vector or trace operater
   >>> l2norm = <psi|psi>
   >>> l2morm = √tr(Aˆ†*A)
   """
   if typex(x) != 'oper':
      x = operx(x) #get oper x
   return np.sqrt(tracex(dotx(daggx(x),x)))

def normx(x):
   """
   to normalize x
   use the Frobeius norm
   >>> |psi>/sqrt(<psi|psi>) # for vector states
   >>> (Aˆ†*A)/tr(Aˆ†*A) # for oper
   """
   if typex(x) != 'oper':
      return x/np.sqrt(l2normx(x))
   else:
      if x.shape[0] == x.shape[1]:
         return dotx(daggx(x),x)/l2normx(x)**2
      else:
         raise TypeError('not a square matrix')

# -----------
# hidden defs
from collections.abc import Sequence
def get_shape(lst, shape=()):
   """
   returns the shape of nested lists similarly to numpy's shape.

   :param lst: the nested list
   :param shape: the shape up to the current recursion depth
   :return: the shape including the current depth
            (finally this will be the full depth)
   """

   if not isinstance(lst, Sequence):
      # base case
      return shape

   # peek ahead and assure all lists in the next depth
   # have the same length
   if isinstance(lst[0], Sequence):
      l = len(lst[0])
      if not all(len(item) == l for item in lst):
         msg = 'not all lists have the same length'
         raise ValueError(msg)

   shape += (len(lst), )

   # recurse
   shape = get_shape(lst[0], shape)

   return shape




# -*- coding: utf-8 -*-
""" usefull code for barren plateau
"""

__all__ = ['plateau']


import qiskit
import tqix.vqa.constants
import tqix.vqa.circuits
import copy

import numpy as np


def plateau(qc: qiskit.QuantumCircuit, 
            qcirs,
            cost_func,
            num_samples):
            
    """ to run plateau
    
    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - cirs (Function): circuit functions
        - coefs (Numpy array): coefs
        - params (Numpy array): Parameters
        - cost_func: name of cost function
        - num_sample = number of sample

    Returns:
        - barren plateau
    """
    
    #num_qubits = qc.num_qubits
    grads = []
    
    for _ in range(num_samples):
        #need new params here
        param_len = len(qcirs[0][2])
        qcirs[0][2] = np.random.uniform(0, 2 * np.pi,param_len)
        qcirs[-1][2] = np.random.uniform(0, 2 * np.pi,param_len)
        
        grad = gradf(qc.copy(),qcirs,cost_func)      
        grads.append(grad)
    aveg = np.average(grads)     
    var =  np.var(grads)      
    return np.abs(aveg), var


def gradf(qc: qiskit.QuantumCircuit,qcirs,cost_func):
    
    """Return the gradient of the loss function
        w.r.t theta_1

    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - cirs (Function): circuit functions
        - coefs (Numpy array): coefs
        - params (Numpy array): Parameters
        - cost_func: name of cost function

    Returns:
        - gradient of loss function w.r.t. theta_0 
    """
    
    s = tqix.vqa.constants.step_size    
    
    qcirs1, qcirs2 = copy.deepcopy(qcirs), copy.deepcopy(qcirs)
    qcirs1[0][2][0] += s #[0] ansatz_name, [2] ansatz_params [0] first_param
    qcirs2[0][2][0] -= s
    
    cost1 = cost_func(qc.copy(),qcirs1)
    cost2 = cost_func(qc.copy(),qcirs2)
    
    return (cost1 - cost2)/(2*s)
    
    
    
    


#!/usr/bin/env python
"""
tqix: a Toolbox for Quantum in X:
    x: quantum measurement, quantum metrology, quantum tomography, and more.
"""
import os

from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# all information about tqix is here
MAN = 3
SUB = 0
SUBS= 4
VERSION = '%d.%d.%d' % (MAN,SUB,SUBS)
REQUIRES = ['numpy (>=1.8)', 'torch (>=1.10)','scipy (>=1.7.1)','sklearn','qiskit (>=0.43.0)']
PACKAGES = ['tqix', 'tqix/dsm', 'tqix/pis', 'tqix/povm','tqix/vqa']

NAME = "tqix"
AUTHOR = ("Binho Le")
AUTHOR_EMAIL = ("binho@fris.tohoku.ac.jp")
LICENSE = "GNU"
DESCRIPTION = "A Toolbox for Quantum in X"
KEYWORDS = "quantum measurement, quantum metrology, quantum tomography"
URL = "https://vqisinfo.wixsite.com/tqix"
PLATFORMS = ["Linux", "Mac OSX", "Unix", "Windows"]

def write_version_py(filename='tqix/version.py'):
    cnt = """\
#this file is generated from tqix setup.py
version = '%(version)s'
"""
    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION})
    finally:
        a.close()

# if exists then remove
if os.path.exists('tqix/version.py'):
    os.remove('tqix/version.py')



write_version_py()

setup(name = NAME,
        version = VERSION,
        description=DESCRIPTION,
        url=URL,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        license=LICENSE,
        packages=PACKAGES,
        long_description=long_description,
        long_description_content_type="text/markdown",
        python_requires=">=3.6",
)


#find . | grep -E "(__pycache__|\.pyc|\.pyo$)" | xargs rm -rf



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: Quangtuan Kieu
>>> all rights reserved
________________________________
"""

from tqix.povm.povm import *
#from tqix.povm.mub3 import *
#from tqix.povm.mub4 import *
#from tqix.povm.mub7 import *

#from tqix.povm.sic2 import *
#from tqix.povm.sic3 import *
#from tqix.povm.sic4 import *
#from tqix.povm.sic5 import *




"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

__all__ = ['cdf','mc']

import numpy as np

def cdf(f):
    """ comulative distribution function
    input: a function f
    return: average of original function

    """
    dim = np.int(len(f))
    nbin = 100
    binh,binb = np.histogram(f,nbin)
    
    s = 0.0 
    for i in range(np.int(nbin/2)):
         s += binh[i] 
    return s/(np.amax(f)/nbin*dim)/(nbin/2.)

from tqix.utility import randunit
def mc(f,niter = 1000):
    """
    Markov chain Monte Carlo simulation
    input: a probability
    return: average of original function
    """
    count = 0
    for i in range (niter):
        if randunit() <= f:
           count += 1
    return count/float(niter)



# -*- coding: utf-8 -*-
""" Everything belong to circuit
"""

#__all__ = []

import qiskit
import tqix
import numpy as np
from qiskit.circuit.library import GMS

#
# ansatz
#

def create_ansatz(qc: qiskit.QuantumCircuit, 
                    qc_func,
                    params, 
                    num_layers):
    
    """ create an ansatz with a given name
    
    Args:
        - qc: initial circuit
        - qc_func: circuit function (3 agrs)
        - params: number of params
        - num_layers: num layers
    
    Return:
        - quantum circuit
    """    
    
    return qc_func(qc.copy(), num_layers, params)


def ghz_ansatz(
    qc: qiskit.QuantumCircuit,
    theta: float = np.pi / 2,
    **kwargs):
    
    """Create GHZ state with a parameter

    Args:
        - qc (QuantumCircuit): Init circuit
        - theta (Float): Parameter
        - **kwargs: for other parameters

    Returns:
        - QuantumCircuit: the added circuit
    """
    
    if isinstance(theta, float) != True:
        theta = (theta['theta'])
    qc.ry(theta, 0)
    
    qc.h(0)
    for i in range(0, qc.num_qubits - 1):
        qc.cnot(0, i + 1)
    return qc

    
def star_ansatz(
    qc: qiskit.QuantumCircuit,
    num_layers: int,
    params: np.ndarray):
    
    """Create star graph ansatz

    Args:
        qc (qiskit.QuantumCircuit): init circuit
        params (np.ndarray): params

    Returns:
        (qiskit.QuantumCircuit): init circuit
    """
    
    n = qc.num_qubits
    
    if len(params) != num_layers*(2*n - 2):
        raise ValueError(
            'The number of parameter must be num_layers*(2*n - 2)')

    j = 0
    for l in range(0, num_layers, 1):
        for i in range(0, n):
            qc.ry(params[j], i)
            j += 1
        qc.cz(0, 1)
        for i in range(2, n):
            qc.ry(params[j], 0)
            j += 1
            qc.cz(0, i)
        qc.barrier()    
    return qc


def star_ansatz_inv(
    qc: qiskit.QuantumCircuit,
    num_layers: int,
    params: np.ndarray):
    
    """Create invert star graph ansatz

    Args:
        qc (qiskit.QuantumCircuit): init circuit
        params (np.ndarray): params

    Returns:
        (qiskit.QuantumCircuit): init circuit
    """
    return star_ansatz(qc,num_layers,params).inverse()

    
def ring_ansatz(
    qc: qiskit.QuantumCircuit,
    num_layers: int,
    params: np.ndarray):
    
    """Create ring graph ansatz

    Args:
        qc (qiskit.QuantumCircuit): init circuit
        thetas (np.ndarray): params

    Returns:
        (qiskit.QuantumCircuit): init circuit
    """

    n = qc.num_qubits
    if len(params) != num_layers*(2*n):
        raise ValueError(
            'The number of parameter must be num_layers*(2*n)')

    j = 0
    for l in range(0, num_layers, 1):
        for i in range(0, n):
            qc.ry(params[j], i)
            j += 1
        for i in range(0, n - 1, 2):
            qc.cz(i, i + 1)
        if n % 2 == 1:
            for i in range(0, n - 1):
                qc.ry(params[j], i)
                j += 1
        else:
            for i in range(0, n):
                qc.ry(params[j], i)
                j += 1
        for i in range(1, n - 1, 2):
            qc.cz(i, i + 1)
        if n % 2 == 1:
            qc.ry(params[j], n - 1)
            j += 1
        qc.cz(0, n - 1)
        qc.barrier()
    return qc
    
def ring_ansatz_inv(
    qc: qiskit.QuantumCircuit,
    num_layers: int,
    params: np.ndarray):
    
    """Create inverse ring graph ansatz

    Args:
        qc (qiskit.QuantumCircuit): init circuit
        thetas (np.ndarray): params

    Returns:
        (qiskit.QuantumCircuit): init circuit
    """

    return ring_ansatz(qc,num_layers,params).inverse()

    
def squeezing_ansatz(
    qc: qiskit.QuantumCircuit,
    num_layers: int,
    params: np.ndarray):
    
    """Create squeezing ansatz

    Args:
        qc (qiskit.QuantumCircuit): init circuit
        thetas (np.ndarray): params

    Returns:
        (qiskit.QuantumCircuit): init circuit
    """ 
    
    n = qc.num_qubits
    if len(params) != num_layers*n*(n+1):
        raise ValueError(
            'The number of parameter must be num_layers*n*(n+1)')
        
    k = 0
    for l in range(0, num_layers, 1):
        for i in range(0, n):
            qc.ry(params[k], i)
            k += 1
            
        # GMS_z gate
        for i in range(n):
            for j in range(i+1, n):
                qc.rzz(params[k], i, j)
                k += 1
        # RX gate
        for i in range(0, n):
            qc.rx(params[k], i)
            k += 1

        # GMS_x gate
        for i in range(n):
            for j in range(i+1, n):
                qc.rxx(params[k], i, j)   
                k += 1
        qc.barrier()    
    return qc


def squeezing_ansatz_inv(
    qc: qiskit.QuantumCircuit,
    num_layers: int,
    params: np.ndarray):
    
    """Create squeezing ansatz

    Args:
        qc (qiskit.QuantumCircuit): init circuit
        thetas (np.ndarray): params

    Returns:
        (qiskit.QuantumCircuit): init circuit
    """ 
    
    return squeezing_ansatz(qc,num_layers,params).inverse()

    
#@np.vectorize
def create_num_params(qc_name, num_qubits, num_layers):
    """ to create a list of number of parameters:
                    
    Args:
        - qc_name: qc name
        - num_qubtis: number of qubits
        - num_layers: list of layers for these ansatzes

    Returns:
        - total number of paramaters
    """
    
    if qc_name == 'star':
        num_params = num_layers*(2*num_qubits - 2)
    elif qc_name == 'ring':
        num_params = num_layers*2*num_qubits
    elif qc_name == 'squeezing':
        num_params = num_layers*num_qubits*(num_qubits+1)  
    #elif qc_name == 'u_phase':
    #    num_params = 3 # fixed
    else:
        #num_params = 1 # one params for y (noise)
        print('add by yourself')  
    
    return int(num_params)
    

def create_params(cirs,coefs,num_qubits,values):
    
    """ to create a list of parameters:
    
    Args:
        - qc_funcs: list of qc_func, i.e., [create_stargraph_ansatz,create_ringgraph_ansatz]
        - num_layers: list of layers for these ansatzes
        - num_qubtis: number of qubits
        - value: initial value
        
    Returns:
        - list of paramaters [[],[],...]
    """
    
    num_params = create_num_params(cirs, coefs, num_qubits)
    params = []
    
    for i in range(len(values)):
        if values[i] == -1: #quy uoc -1 means random
            params.append(np.random.uniform(0, 2 * np.pi, num_params[i]))    
        else:
            params.append(values[i] * num_params[i])
                
    #add phase to params
    list2str = list(map(lambda f: f.__name__, cirs)) 
    if 'u_phase' in list2str:
        idx = list2str.index('u_phase')  
        params[idx] = values[idx]
    
    return params 


#
# quantum states
#

def state_vector(qc: qiskit.QuantumCircuit):
    # to get the state vector from qc  
    return qiskit.quantum_info.Statevector.from_instruction(qc.copy()).data


def state_density(qc: qiskit.QuantumCircuit):
    # to get the state densiry from qc
    return qiskit.quantum_info.DensityMatrix.from_instruction(qc.copy()).data


def ghz_cir(qc: qiskit.QuantumCircuit, *args, **kwargs):
    
    """Create GHZ state 

    Args:
        - qc (QuantumCircuit): Init circuit
        - *args, **kwargs: nothing

    Returns:
        - QuantumCircuit: the added circuit
    """

    qc.h(0)
    
    for i in range(0, qc.num_qubits - 1):
        qc.cnot(0, i + 1)
        
    return qc


def star_graph(qc: qiskit.QuantumCircuit, *args, **kwargs):
    
    """Create start graph 

    Args:
        - qc (QuantumCircuit): Init circuit
        - *args, **kwargs: None

    Returns:
        - QuantumCircuit: the added circuit
    """
    
    num_qubits = qc.num_qubits
    for i in range(0, num_qubits):
        qc.h(i)
    for i in range(1, num_qubits):
        qc.cz(0, i)
    qc.barrier()   
    
    return qc

def ring_graph(qc: qiskit.QuantumCircuit, *args, **kwargs):
    
    n = qc.num_qubits

    for i in range(0, n - 1, 2):
        qc.cz(i, i + 1)
    if n % 2 == 1:
        for i in range(1, n - 1, 2):
            qc.cz(i, i + 1)
        qc.cz(0, n - 1)
        qc.barrier()
    return qc

def w_cir(qc, *args, **kwargs):
    """Create n-qubit W state based on the its number of qubits

    Args:
        - qc (qiskit.QuantumCircuit): init circuit

    Returns:
        - qiskit.QuantumCircuit
    """
    
    qc.x(0)
    qc = w_sub(qc, qc.num_qubits)
    return qc


def w3(circuit: qiskit.QuantumCircuit, qubit: int):
    """Create W state for 3 qubits

    Args:
        - circuit (qiskit.QuantumCircuit): added circuit
        - qubit (int): the index that w3 circuit acts on

    Returns:
        - qiskit.QuantumCircuit: added circuit
    """
    qc = qiskit.QuantumCircuit(3)
    theta = np.arccos(1 / np.sqrt(3))
    qc.cf(theta, 0, 1)
    qc.cx(1, 0)
    qc.ch(1, 2)
    qc.cx(2, 1)
    w3 = qc.to_gate(label='w3')
    # Add the gate to your circuit which is passed as the first argument to cf function:
    circuit.append(w3, [qubit, qubit + 1, qubit + 2])
    return circuit

def w_sub(qc: qiskit.QuantumCircuit, num_qubits: int, shift: int = 0):
    """The below codes is implemented from [this paper](https://arxiv.org/abs/1606.09290)
    \n Simplest case: 3 qubits. <img src='../images/general_w.png' width = 500px/>
    \n General case: more qubits. <img src='../images/general_w2.png' width = 500px/>

    Args:
        - num_qubits (int): number of qubits
        - shift (int, optional): begin wire. Defaults to 0.

    Raises:
        - ValueError: When the number of qubits is not valid

    Returns:
        - qiskit.QuantumCircuit
    """
    #num_qubits = qc.num_qubits
    
    if num_qubits < 2:
        raise ValueError('W state must has at least 2-qubit')
    if num_qubits == 2:
        # |W> state ~ |+> state
        qc.h(0)
        return qc
    if num_qubits == 3:
        # Return the base function
        qc.w3(shift)
        return qc
    else:
        # Theta value of F gate base on the circuit that it acts on
        theta = np.arccos(1 / np.sqrt(qc.num_qubits - shift))
        qc.cf(theta, shift, shift + 1)
        # Recursion until the number of qubits equal 3
        w_sub(qc, num_qubits - 1, qc.num_qubits - (num_qubits - 1))
        for i in range(1, num_qubits):
            qc.cnot(i + shift, shift)
    return qc

def w_inv(qc: qiskit.QuantumCircuit):
    """Create n-qubit W state based on the its number of qubits

    Args:
        - qc (qiskit.QuantumCircuit): init circuit

    Returns:
        - qiskit.QuantumCircuit
    """
    qc1 = qiskit.QuantumCircuit(qc.num_qubits)
    qc1.x(0)
    qc1 = w_sub(qc1, qc.num_qubits)
    qc = qc.combine(qc1.inverse())
    return qc


def cf(qc: qiskit.QuantumCircuit, theta: float, qubit1: int, qubit2: int):
    """Add Controlled-F gate to quantum circuit

    Args:
        - qc (qiskit.QuantumCircuit): ddded circuit
        - theta (float): arccos(1/sqrt(num_qubits), base on number of qubit
        - qubit1 (int): control qubit
        - qubit2 (int): target qubit

    Returns:
        - qiskit.QuantumCircuit: Added circuit
    """
    cf = qiskit.QuantumCircuit(2)
    u = np.array([[1, 0, 0, 0], [0, np.cos(theta), 0,
                            np.sin(theta)], [0, 0, 1, 0],
                [0, np.sin(theta), 0, -np.cos(theta)]])
    cf.unitary(u, [0, 1])
    cf_gate = cf.to_gate(label='CF')
    qc.append(cf_gate, [qubit1, qubit2])
    return qc

qiskit.QuantumCircuit.w3 = w3
qiskit.QuantumCircuit.cf = cf


#
# measurements
#

def measure(qc: qiskit.QuantumCircuit, qubits, cbits = []):
    """Measuring the quantum circuit which fully measurement gates

    Args:
        - qc (QuantumCircuit): Measured circuit
        - qubits (Numpy array): List of measured qubit

    Returns:
        - float: Frequency of 00..0 cbit
    """    
    n = len(qubits)
    if cbits == []:
        cbits = qubits.copy()
    for i in range(0, n):
        qc.measure(qubits[i], cbits[i])
    
    counts = qiskit.execute(
            qc, backend = tqix.vqa.constants.backend,
            shots = tqix.vqa.constants.num_shots).result().get_counts()

    return counts.get("0" * len(qubits), 0) / tqix.vqa.constants.num_shots


def measure_theor(qc: qiskit.QuantumCircuit, qubits, cbits = []):
    """Measuring the quantum circuit which fully measurement gates

    Args:
        - qc (QuantumCircuit): Measured circuit
        - qubits (Numpy array): List of measured qubit

    Returns:
        - float: Frequency of 00..0 cbit by Born rule
    """    
    rho = qiskit.quantum_info.DensityMatrix.from_instruction(qc.copy()).data
    pro0 = np.real(rho[0,0])
    return pro0


def measure_all(qc: qiskit.QuantumCircuit):
    """Measuring the quantum circuit which fully measurement gates

    Args:
        - qc (QuantumCircuit): Measured circuit
        - qubits (Numpy array): List of measured qubit

    Returns:
        - float: all cbits
    """
    
    n = qc.num_qubits
    qubits = qc.qubits
    cbits = qc.clbits
    for i in range(0, n):
        qc.measure(qubits[i], cbits[i])
    #qc.measure_all()

    shots = tqix.vqa.constants.num_shots
    counts = qiskit.execute(
            qc, backend= tqix.vqa.constants.backend,
            shots=shots).result().get_counts()

    new_counts = dict()
    for i in range(2**n):
        bin_str = bin(i)[2:]
        if len(bin_str) < n:
            bin_str = "0" * (n - len(bin_str)) + bin_str
        new_counts[bin_str] = counts.get(bin_str, 0)
    
    # calculate the probabilities for each bit value
    probs = {}
    for output in new_counts:
        probs[output] = new_counts[output]/(1.0*shots)
    return np.array(list(probs.values()))


def measure_born(qc: qiskit.QuantumCircuit):
    """Measuring the quantum circuit which fully measurement gates

    Args:
        - qc (QuantumCircuit): Measured circuit
        - qubits (Numpy array): List of measured qubit

    Returns:
        - float: all cbits
    """
    
    n = qc.num_qubits
    d = 2**n
    probs = []

    rho = qiskit.quantum_info.DensityMatrix.from_instruction(qc.copy()).data

    for i in range(0, d):
        probs.append(np.real(rho[i,i]))

    return  np.array(probs) 




import numpy as np
from scipy.sparse import csc_matrix
import torch 

__all__ = ['Sx','Sy','Sz','S_plus','S_minus','S_2']

def Sx(S,use_gpu = False,device='cuda'):
    """

    :param S: dimension of spin operator
    :type S: int
    :param use_gpu: if use gpu, defaults to False
    :type use_gpu: bool, optional
    :param device: name of compute device, defaults to 'cuda'
    :type device: str, optional
    :return: spin operator
    :rtype: ndarray, tensor, sparse
    """    
    dim = int(2*S+1)
    if use_gpu:
        m,m_prime = torch.arange(-S,S+1).flip([0]).to(device),torch.arange(-S,S+1).flip([0]).to(device)
        m_ind,m_prime_ind = torch.meshgrid(torch.arange(dim).to(device), torch.arange(dim).to(device),indexing='ij')
        m_prime_ind_plus = m_prime_ind+1
        m_ind_plus = m_ind+1
        non_zero_row_inds,non_zero_col_inds =  torch.nonzero((m_ind_plus == m_prime_ind).type(torch.int64)+(m_ind == m_prime_ind_plus).type(torch.int64),as_tuple=True)
        coor = torch.stack([non_zero_row_inds,non_zero_col_inds])
        non_zero_values = 1/2*torch.sqrt(S*(S+1)-torch.mul(m[non_zero_row_inds],m_prime[non_zero_col_inds]))
        opr = torch.sparse_coo_tensor(coor, non_zero_values, (dim, dim)).to_dense()
    else: 
        m,m_prime = np.arange(-S,S+1)[::-1],np.arange(-S,S+1)[::-1]
        m_ind,m_prime_ind = np.meshgrid(np.arange(dim), np.arange(dim), indexing='ij')
        m_prime_ind_plus = m_prime_ind+1
        m_ind_plus = m_ind+1
        non_zero_row_inds,non_zero_col_inds =  np.nonzero((m_ind_plus == m_prime_ind).astype(np.int64)+(m_ind == m_prime_ind_plus).astype(np.int64))
        non_zero_values = 1/2*np.sqrt(S*(S+1)-np.multiply(m[non_zero_row_inds],m_prime[non_zero_col_inds]))
        opr = csc_matrix((non_zero_values, (non_zero_row_inds,non_zero_col_inds)), shape=(dim, dim))
    return opr

def Sy(S,use_gpu = False,device='cuda'):
    """

    :param S: dimension of spin operator
    :type S: int
    :param use_gpu: if use gpu, defaults to False
    :type use_gpu: bool, optional
    :param device: name of compute device, defaults to 'cuda'
    :type device: str, optional
    :return: spin operator
    :rtype: ndarray, tensor, sparse
    """    
    dim = int(2*S+1)
    if use_gpu:
        m,m_prime = torch.arange(-S,S+1).flip([0]).to(device),torch.arange(-S,S+1).flip([0]).to(device)
        m_ind,m_prime_ind = torch.meshgrid(torch.arange(dim).to(device), torch.arange(dim).to(device),indexing='ij')
        m_prime_ind_plus = m_prime_ind+1
        m_ind_plus = m_ind+1
        kron_delta = (m_prime_ind == m_ind_plus).type(torch.int64) - (m_prime_ind_plus == m_ind).type(torch.int64) 
        non_zero_row_inds,non_zero_col_inds = torch.nonzero(kron_delta,as_tuple=True)
        coor = torch.stack([non_zero_row_inds,non_zero_col_inds])
        non_zero_kron_delta = kron_delta[non_zero_row_inds,non_zero_col_inds]
        non_zero_values = non_zero_kron_delta*1/(2j)*torch.sqrt(S*(S+1)-torch.mul(m[non_zero_row_inds],m_prime[non_zero_col_inds]))
        opr = torch.sparse_coo_tensor(coor, non_zero_values, (dim, dim)).to_dense()
    else:
        m,m_prime = np.arange(-S,S+1)[::-1],np.arange(-S,S+1)[::-1]
        m_ind,m_prime_ind = np.meshgrid(np.arange(dim), np.arange(dim), indexing='ij')
        m_prime_ind_plus = m_prime_ind+1
        m_ind_plus = m_ind+1
        kron_delta = (m_prime_ind == m_ind_plus).astype(np.int64) - (m_prime_ind_plus == m_ind).astype(np.int64) 
        non_zero_row_inds,non_zero_col_inds = np.nonzero(kron_delta)
        non_zero_kron_delta = kron_delta[non_zero_row_inds,non_zero_col_inds]
        non_zero_values = non_zero_kron_delta*np.array(1/(2j))*np.sqrt(S*(S+1)-np.multiply(m[non_zero_row_inds],m_prime[non_zero_col_inds]))
        opr = csc_matrix((non_zero_values, (non_zero_row_inds,non_zero_col_inds)), shape=(dim, dim))
    return opr

def Sz(S,use_gpu = False,device='cuda'):
    """

    :param S: dimension of spin operator
    :type S: int
    :param use_gpu: if use gpu, defaults to False
    :type use_gpu: bool, optional
    :param device: name of compute device, defaults to 'cuda'
    :type device: str, optional
    :return: spin operator
    :rtype: ndarray, tensor, sparse
    """    
    dim = int(2*S+1)
    if use_gpu:
        m,_ = torch.arange(-S,S+1).flip([0]).to(device),torch.arange(-S,S+1).flip([0]).to(device)
        m_ind,m_prime_ind = torch.meshgrid(torch.arange(dim).to(device), torch.arange(dim).to(device), indexing='ij')
        kron_delta = (m_ind == m_prime_ind).type(torch.int64) 
        non_zero_row_inds,non_zero_col_inds = torch.nonzero(kron_delta,as_tuple=True)
        non_zero_kron_delta = kron_delta[non_zero_row_inds,non_zero_col_inds]
        non_zero_values = non_zero_kron_delta*m[non_zero_row_inds]
        coor = torch.stack([non_zero_row_inds,non_zero_col_inds])
        opr = torch.sparse_coo_tensor(coor, non_zero_values, (dim, dim)).to_dense()
    else:
        m,_ = np.arange(-S,S+1)[::-1],np.arange(-S,S+1)[::-1]
        m_ind,m_prime_ind = np.meshgrid(np.arange(dim), np.arange(dim), indexing='ij')
        kron_delta = (m_ind == m_prime_ind).astype(np.int64) 
        non_zero_row_inds,non_zero_col_inds = np.nonzero(kron_delta)
        non_zero_kron_delta = kron_delta[non_zero_row_inds,non_zero_col_inds]
        non_zero_values = non_zero_kron_delta*m[non_zero_row_inds]
        opr = csc_matrix((non_zero_values, (non_zero_row_inds,non_zero_col_inds)), shape=(dim, dim))
    return opr

def S_minus(S,use_gpu = False,device='cuda'):
    """

    :param S: dimension of spin operator
    :type S: int
    :param use_gpu: if use gpu, defaults to False
    :type use_gpu: bool, optional
    :param device: name of compute device, defaults to 'cuda'
    :type device: str, optional
    :return: spin operator
    :rtype: ndarray, tensor, sparse
    """      
    dim = int(2*S+1)
    if use_gpu:
        m,m_prime = torch.arange(-S,S+1).flip([0]).to(device),torch.arange(-S,S+1).flip([0]).to(device)
        m_ind,m_prime_ind = torch.meshgrid(torch.arange(dim).to(device), torch.arange(dim).to(device), indexing='ij')
        m_prime_ind_plus = m_prime_ind+1
        non_zero_row_inds,non_zero_col_inds = torch.nonzero((m_prime_ind_plus==m_ind).type(torch.int64),as_tuple=True)
        non_zero_values = torch.sqrt(S*(S+1)-torch.mul(m[non_zero_row_inds],m_prime[non_zero_col_inds]))
        coor = torch.stack([non_zero_row_inds,non_zero_col_inds])
        opr = torch.sparse_coo_tensor(coor, non_zero_values, (dim, dim)).to_dense()
    else:
        m,m_prime = np.arange(-S,S+1)[::-1],np.arange(-S,S+1)[::-1]
        m_ind,m_prime_ind = np.meshgrid(np.arange(dim), np.arange(dim), indexing='ij')
        m_prime_ind_plus = m_prime_ind+1
        non_zero_row_inds,non_zero_col_inds = np.nonzero((m_prime_ind_plus==m_ind).astype(np.int64))
        non_zero_values = np.sqrt(S*(S+1)-np.multiply(m[non_zero_row_inds],m_prime[non_zero_col_inds]))
        opr = csc_matrix((non_zero_values, (non_zero_row_inds,non_zero_col_inds)), shape=(dim, dim))
    return opr

def S_plus(S,use_gpu = False,device='cuda'):
    """

    :param S: dimension of spin operator
    :type S: int
    :param use_gpu: if use gpu, defaults to False
    :type use_gpu: bool, optional
    :param device: name of compute device, defaults to 'cuda'
    :type device: str, optional
    :return: spin operator
    :rtype: ndarray, tensor, sparse
    """    
    dim = int(2*S+1)
    if use_gpu:
        m,m_prime = torch.arange(-S,S+1).flip([0]).to(device),torch.arange(-S,S+1).flip([0]).to(device)
        m_ind,m_prime_ind = torch.meshgrid(torch.arange(dim).to(device), torch.arange(dim).to(device), indexing='ij')
        m_ind_plus = m_ind+1
        non_zero_row_inds,non_zero_col_inds =  torch.nonzero((m_prime_ind==m_ind_plus).type(torch.int64),as_tuple=True)
        non_zero_values = torch.sqrt(S*(S+1)-torch.mul(m[non_zero_row_inds],m_prime[non_zero_col_inds]))
        coor = torch.stack([non_zero_row_inds,non_zero_col_inds])
        opr = torch.sparse_coo_tensor(coor, non_zero_values, (dim, dim)).to_dense()
    else:
        m,m_prime = np.arange(-S,S+1)[::-1],np.arange(-S,S+1)[::-1]
        m_ind,m_prime_ind = np.meshgrid(np.arange(dim), np.arange(dim), indexing='ij')
        m_ind_plus = m_ind+1
        non_zero_row_inds,non_zero_col_inds = np.nonzero((m_prime_ind == m_ind_plus).astype(np.int64))
        non_zero_values = np.sqrt(S*(S+1)-np.multiply(m[non_zero_row_inds],m_prime[non_zero_col_inds]))
        opr = csc_matrix((non_zero_values, (non_zero_row_inds,non_zero_col_inds)), shape=(dim, dim))
    return opr

def S_2(S):
    """

    :param S: dimension of spin operator
    :type S: int
    :return: opr: spin operator
    :rtype: ndarray, tensor, sparse
    """    
    dim = int(2*S+1)
    m_ind,m_prime_ind = np.meshgrid(np.arange(dim), np.arange(dim), indexing='ij')
    non_zero_row_inds,non_zero_col_inds = np.nonzero((m_ind == m_prime_ind).astype(np.int64))
    non_zero_values = np.array([S*(S+1)]*len(non_zero_row_inds))
    opr = csc_matrix((non_zero_values, (non_zero_row_inds,non_zero_col_inds)), shape=(dim, dim))
    return opr


# -*- coding: utf-8 -*-
""" Variational quantum metrology
    Author: Le Bin Ho @ 2023
"""

#__all__ = []

import qiskit
import numpy as np
import tqix.vqa.fitting 

from qiskit.extensions import UnitaryGate
from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit

#model can dua params vao

def training(qc: qiskit.QuantumCircuit,
            qcirs,
            which_train,
            cost_func,
            grad_func,
            opt_func,
            num_steps):
    
    """ train a vqm model
    
    Args:
        - qc (QuantumCircuit): Fitting circuit
        - qcirs: set of circuits functions
        - which_train: training option
        - coss_func (FunctionType): coss function
        - grad_func (FunctionType): Gradient function
        - opt_func (FunctionType): Otimizer function
        - num_steps (Int): number of iterations
        
    Returns:
        - optimal parameters
    """
    

    return tqix.vqa.fitting.fit(qc,qcirs,which_train,cost_func,grad_func,opt_func,num_steps)
    

def qc_add(qc, qcirs):
    
    """ create a full circuit from qc_func
    
    Args:
        - qc: initial circuit
        - qcirs: list of circuits, num_layers, parameters, ....
        
    Returns:
        - qc: final circuit
    """
    
    cirq = qc.copy()    
    for i in range (len(qcirs)):
        cirq &= qcirs[i][0](qc.copy(), qcirs[i][1], qcirs[i][2])
        
    return cirq

#
# custom unitary for phase and noises
#

def u_phase(qc: qiskit.QuantumCircuit, t, params):
    
    """Add phase model to the circuit
    
    Args:
        - qc (qiskit.QuantumCircuit): quantumcircuit
        - t: time coefs
        - params  (number of parameters)
        
    Return
        - qc
    """
    n = qc.num_qubits
    
    for i in range(0,n):     
        qc.append(_u_gate(t, params),[i])
    
    return qc

def u_phase_anc(qc: qiskit.QuantumCircuit, t, params):
    
    """Add phase model to the circuit from anc
    
    Args:
        - qc (qiskit.QuantumCircuit): quantumcircuit
        - t: time coefs
        - params  (number of parameters)
        - a, b (int): start and end
        
    Return
        - qc
    """
    n = qc.num_qubits
    
    for i in range(1,n):     
        qc.append(_u_gate(t, params),[i])
    
    return qc

def _u_gate(t, params):
    """ return arbitary gates
    
    Args:
        - t: time coefs
        - params: phases
    
    Method:
        H = x*sigma_x + y*sigma_y +  z*sigma_z
        U = exp(-i*t*H)
        t: time (use for time-dephasing)
    """
        
    x,y,z = params[0],params[1],params[2]
    p2 = np.sqrt(x*x + y*y + z*z)
    tp2 = t * p2
    
    u11 = np.cos(tp2) - 1j*z*np.sin(tp2)/p2
    u12 = (-1j*x - y)*np.sin(tp2)/p2
    u21 = (-1j*x + y)*np.sin(tp2)/p2
    u22 = np.cos(tp2) + 1j*z*np.sin(tp2)/p2
    
    u = [[u11, u12],[u21,u22]]
    gate = UnitaryGate(u, 'u_phase')

    return gate



# -*- coding: utf-8 -*-
""" usefull base
"""

#__all__ = ['def_cost_func',
#           'grad_cost']


import qiskit
import tqix.vqa.constants
import tqix.vqa.bounds

import numpy as np
from numpy.linalg import inv
import copy

def trace_f_invq(qc, qcirs):                        
    """Return the cost function
        C = 1 - tr[cfim*qfim.inv]/d
    
    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: set of circuits 

    Returns:
        - Numpy array: The cost function
    """
        
    #calculate F and Q here
    qfim = tqix.vqa.bounds.sld_qfim(qc.copy(),qcirs)          
    cfim = tqix.vqa.bounds.cfim(qc.copy(),qcirs) 
    inv_qfim = inv(qfim + np.eye(len(qfim)) * 10e-10)
    
    return 1 - np.trace(cfim @ inv_qfim)/len(qcirs[1][2]) #[1]:u_phase,[1][2]:phases


def bound_sld_cls(qc, qcirs):
    """Return the cost function
        C = 1 - sld_bound/cls_bound
    
    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: set of circuits 

    Returns:
        - Numpy array: The cost function
    """
    
    sld_bound = tqix.vqa.bounds.sld_bound(qc.copy(),qcirs) 
    rld_bound = tqix.vqa.bounds.rld_bound(qc.copy(),qcirs) 
    cls_bound = tqix.vqa.bounds.cls_bound(qc.copy(),qcirs)

    return 1 - sld_bound/cls_bound


def bound_rld_cls(qc, qcirs):           
    """Return the cost function
        C = 1 - sld_bound/cls_bound
    
    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: set of circuits 

    Returns:
        - Numpy array: The cost function
    """
    
    rld_bound = tqix.vqa.bounds.rld_bound(qc.copy(),qcirs)  
    cls_bound = tqix.vqa.bounds.cls_bound(qc.copy(),qcirs)
    
    return 1 - rld_bound/cls_bound


def norm2(qc, qcirs):
    """Return the cost function
        C = trace(f) - trace(q)
    
    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: set of circuits 

    Returns:
        - Numpy array: The cost function
    """
    #calculate F and Q here
    qfim = tqix.vqa.bounds.sld_qfim(qc.copy(),qcirs)          
    cfim = tqix.vqa.bounds.cfim(qc.copy(),qcirs)
    inv_qfim = inv(qfim + np.eye(len(qfim)) * 10e-10)
    
    return np.linalg.norm(cfim - qfim, ord = 2) #np.trace(inv_qfim) 


def cost_func_sld(qc, qcirs):
    """Return the cost function sld
        C = sld_qfim
    
    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: set of circuits 

    Returns:
        - Numpy array: The cost function
    """
    sld_qfim = tqix.vqa.bounds.sld_bound(qc.copy(),qcirs)     
        
    return sld_qfim


def cost_func_cls(qc, qcirs):
    """Return the cost function cls
        C = cls_qfim
    
    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: set of circuits 

    Returns:
        - Numpy array: The cost function
    """
    cfim = tqix.vqa.bounds.cls_bound(qc.copy(),qcirs)     
        
    return cfim


def grad_cost(qc, qcirs, cost_func, which_train):
    
    """Return the gradient of the loss function
        C = 1 - tr[cfim*qfim.inv]/d
        => nabla_C

    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: set of circuits 
        - cost_func: defined cost function
        - which_train (Numpy array): setting which train

    Returns:
        - Numpy array: The vector of gradient
    """
    
    
    s = tqix.vqa.constants.step_size
    grad_cost = [] #to reset index
        
    for i in which_train:
        for j in range(len(qcirs[i][2])):  #[i]:vi tri cirquit, [2]:param in circuit[i]
            qcirs1, qcirs2 = copy.deepcopy(qcirs), copy.deepcopy(qcirs)
            qcirs1[i][2][j] += s
            qcirs2[i][2][j] -= s

            cost1 = cost_func(qc.copy(),qcirs1)
            cost2 = cost_func(qc.copy(),qcirs2)
            
            grad_cost.append((cost1 - cost2)/(2*s))
            
    return grad_cost

def fit(qc: qiskit.QuantumCircuit,
        qcirs,
        which_train,
        cost_func,
        grad_func,
        opt_func,
        num_steps):
    """Return the new thetas that fit with the circuit from create_circuit_func function#

    Args:
        - qc (QuantumCircuit): Fitting circuit
        - qcirs: set of circuits functions
        - which_train: training option
        - coss_func (FunctionType): coss function
        - grad_func (FunctionType): Gradient function
        - opt_func (FunctionType): Otimizer function
        - num_steps (Int): number of iterations

    Returns:
        - thetas (Numpy array): the optimized parameters
        - loss_values (Numpy array): the list of loss_value
    """
    paramss = []
    costs = []
        
    # get params_train
    params_train = []
    for i in which_train:
        params_train.extend(qcirs[i][2]) #[i]:vi tri cirquit, [2]:param in circuit[i]
    
    for s in range(0, num_steps):
        dev_cost = grad_func(qc.copy(),qcirs,cost_func,which_train)
            
        # calculate params_train
        params_train = opt_func(params_train,dev_cost,s)
        
                
        # update full params
        pre_ind = 0
        for i in which_train:              
            qcirs[i][2] = params_train[pre_ind:pre_ind+len(qcirs[i][2])]
            pre_ind += len(qcirs[i][2])           

        # compute cost   
        cost = cost_func(qc.copy(),qcirs)
        
        costs.append(cost)
        print(s,cost)
        
    return qcirs, costs


def sgd(params, dev_cost, i):
    """optimizer standard gradient descene
    
    Args:
        - params: parameters
        - dev_cost: gradient of the cost function
        - i: use for adam (dont use here)
    
    Returns:
        - params
    """
    learning_rate = tqix.vqa.constants.learning_rate
    params -= learning_rate*dev_cost
    return params
    
def adam(params, dev_cost, iteration):
    """optimizer adam
    
    Args:
        - params: parameters
        - dev_cost: gradient of the cost function
    
    Returns:
        - params
    """    
    num_params = len(params) #.shape[0]
    beta1, beta2,epsilon = 0.8, 0.999, 10**(-8)
    
    m, v = list(np.zeros(num_params)), list(
                    np.zeros(num_params))
    
    for i in range(num_params):
        m[i] = beta1 * m[i] + (1 - beta1) * dev_cost[i]
        v[i] = beta2 * v[i] + (1 - beta1) * dev_cost[i]**2
        mhat = m[i] / (1 - beta1**(iteration + 1))
        vhat = v[i] / (1 - beta2**(iteration + 1))
   
        params[i] -= tqix.vqa.constants.learning_rate * mhat / (np.sqrt(vhat) + epsilon)
    
    return params
    
    
    


# -*- coding: utf-8 -*-
""" Variational quantum metrology
    Author: Le Bin Ho @ 2023
"""

#__all__ = []

import qiskit
import numpy as np

from qiskit.extensions import UnitaryGate
from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit
from qiskit.quantum_info import Kraus
from qiskit.providers.aer.utils import approximate_quantum_error
from qiskit.providers.aer.noise import QuantumError


# various noise via Kraus operators
def dephasing(qc, t, y):
    """Add dephasing to the circuit

    Args:
        - qc (qiskit.QuantumCircuit): quantumcircuit
        - t: time coefs
        - y: gamma in params
    
    Return
        - qc
    """ 
    y = y[0] if type(y) == list else y
    lamb = 1 - np.exp(-t*y)

    # kraus operators
    k1 = np.array([[1, 0],[0, np.sqrt(1 - lamb)]])
    k2 = np.array([[0, 0],[0, np.sqrt(lamb)]])
    
    noise_ops = Kraus([k1,k2])
    kraus_to_error = QuantumError(noise_ops) 
    
    for i in range(qc.num_qubits):
        qc.append(kraus_to_error,[i])
        
    return qc


def bit_flip(qc, t, y):
    """Add bit flip to the circuit

    Args:
        - qc (qiskit.QuantumCircuit): quantumcircuit
        - t: time
        - y: gamma
    
    Return
        - qc
    """ 
    y = y[0] if type(y) == list else y
    lamb = 1 - np.exp(-t*y)
    
    noise_ops = Kraus([np.sqrt(1-lamb) * np.array([[1, 0], [0, 1]]),
                   np.sqrt(lamb) * np.array([[0, 1], [1, 0]])])
    kraus_to_error = QuantumError(noise_ops)
    
    for i in range(qc.num_qubits):
        qc.append(kraus_to_error,[i])
        
    return qc


def phase_flip(qc, t, y):
    """Add phase flip to the circuit

    Args:
        - qc (qiskit.QuantumCircuit): quantumcircuit
        - t: time
        - y: gamma
    
    Return
        - qc
    """ 
    y = y[0] if type(y) == list else y
    lamb = 1 - np.exp(-t*y)
    
    noise_ops = Kraus([np.sqrt(1-lamb) * np.array([[1, 0], [0, 1]]),
                   np.sqrt(lamb) * np.array([[1, 0], [0, -1]])])
    kraus_to_error = QuantumError(noise_ops)
    
    for i in range(qc.num_qubits):
        qc.append(kraus_to_error,[i])
        
    return qc


def depolarizing(qc, t, y):
    """Add depolarizing to the circuit

    Args:
        - qc (qiskit.QuantumCircuit): quantumcircuit
        - t: time
        - y: gamma
    
    Return
        - qc
    """ 
    y = y[0] if type(y) == list else y
    lamb = 1 - np.exp(-t*y)
    
    noise_ops = Kraus([np.sqrt(1-lamb) * np.array([[1, 0], [0, 1]]),
                       np.sqrt(lamb/3.) * np.array([[0, 1], [1, 0]]),
                       np.sqrt(lamb/3.) * np.array([[0, -1j], [1j, 0]]),
                       np.sqrt(lamb/3.) * np.array([[1, 0], [0, -1]])
                        ])
    kraus_to_error = QuantumError(noise_ops)
    
    for i in range(qc.num_qubits):
        qc.append(kraus_to_error,[i])
        
    return qc


def markovian(qc: qiskit.QuantumCircuit, t, y):
    """Add bit flip to the circuit

    Args:
        - qc (qiskit.QuantumCircuit): quantumcircuit
        - t: time
        - y: gamma
    
    Return
        - qc
    """ 
    y = y[0] if type(y) == list else y
    qt = 1 - np.exp(-y*t) 
    
    k1 = np.array([[np.sqrt(1-qt) ,0], [0,1]])
    k2 = np.array([[np.sqrt(qt),0], [0,0]])
    
    noise_ops = Kraus([k1,k2])
    kraus_to_error = QuantumError(noise_ops)  
    
    for i in range(qc.num_qubits):
        qc.append(kraus_to_error,[i])
    
    return qc


def non_markovian(qc: qiskit.QuantumCircuit, t, y):
    """Add bit flip to the circuit

    Args:
        - qc (qiskit.QuantumCircuit): quantumcircuit
        - t: time
        - y: gamma
    
    Return
        - qc
    """ 
    y = y[0] if type(y) == list else y
    tc = 20.0 #fixed in rXiv:2305.08289
    qt = 1 - np.exp(-y*t*t/(2*tc))
    
    k1 = np.array([[np.sqrt(1-qt) ,0], [0,1]])
    k2 = np.array([[np.sqrt(qt),0], [0,0]])
    
    noise_ops = Kraus([k1,k2])
    kraus_to_error = QuantumError(noise_ops)  
    
    for i in range(qc.num_qubits):
        qc.append(kraus_to_error,[i])
    
    return qc


"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: 
>>> all rights reserved
________________________________
"""


from tqix.pis.squeeze_param import *
from tqix.pis.spin_operators import *
from tqix.pis.noise import *
from tqix.pis.gates import *
from tqix.pis.circuit import *
from tqix.pis.optimizers import *
from tqix.pis.util import *


"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: Quangtuan Kieu
>>> all rights reserved
________________________________
"""

from tqix.dsm.dsmStrong import execu as exeStr
from tqix.dsm.dsmWeak import execu as exeWeak
from tqix.dsm.dsmProb import execu as exeProb
from tqix.dsm.dsmProb_Conf_1 import execu as exeProb_1
from tqix.dsm.dsmProb_Conf_2 import execu as exeProb_2

class execute:
   # to run the program
   
   def __init__(self,state,niter,nsamp,theta,m=0.,st=0.,pHist=False):
      self.state = state
      self.niter = niter
      self.nsamp = nsamp
      self.theta = theta
      self.m = m
      self.st = st
      self.pHist = pHist

   def job(self,name):
      if name == 'strong':
         model = exeStr(self.state,\
                  self.niter,self.nsamp,\
                  self.theta,self.pHist)
      if name == 'weak':
         model = exeWeak(self.state,\
                  self.niter,self.nsamp,\
                  self.theta,self.pHist)
      if name == 'prob':
         model = exeProb(self.state,\
                  self.niter,self.nsamp,\
                  self.theta,self.pHist)
      if name == 'prob_1':
         model = exeProb_1(self.state,\
                  self.niter,self.nsamp,\
                  self.theta,self.m,self.st,self.pHist)
      if name == 'prob_2':
         model = exeProb_2(self.state,\
                  self.niter,self.nsamp,\
                  self.theta,self.m,self.st,self.pHist)
      return model

   def info(self):
      print("This is to run a DSM code")



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

__all__ = ['about']

import sys
import os
import platform
import numpy
import scipy
import inspect
import tqix.version
from   tqix.hinfo import hinfo

def about():
    """
    About box for tqix. Gives version numbers for
    QSim, NumPy, SciPy, Cython, and MatPlotLib.
    """
    print("")
    print("tqix: Quantum computor Simulation code")
    print("copyright (c) 2019 and later.")
    print("authors: Binho Le ")
    print("")
    print("tqix Version:       %s" % tqix.__version__)
    print("Numpy Version:      %s" % numpy.__version__)
    print("Scipy Version:      %s" % scipy.__version__)
    try:
        import Cython
        cython_ver = Cython.__version__
    except:
        cython_ver = 'None'
    print("Cython Version:     %s" % cython_ver)
    try:
        import matplotlib
        matplotlib_ver = matplotlib.__version__
    except:
        matplotlib_ver = 'None'
    print("Matplotlib Version: %s" % matplotlib_ver)
    print("Python Version:     %d.%d.%d" % sys.version_info[0:3])
    print("Number of CPUs:     %s" % hinfo()['cpus'])
    print("Platform Info:      %s (%s)" % (platform.system(),
                                           platform.machine()))
    print("")

if __name__ == "__main__":
    about()



from tqix.pis.util import *
from tqix.pis import *
import numpy as np 
import torch

__all__ = ['get_xi_2_H','get_xi_2_S','get_xi_2_R','get_xi_2_D','get_xi_2_E','get_xi_2_F']

def get_xi_2_H(alpha,beta,gamma,qc):
    """

    :param alpha: type of alpha
    :type alpha: str
    :param beta: type of beta
    :type beta: str
    :param gamma: type of gamma
    :type gamma: str
    :param qc: circuit object 
    :type qc: Circuit
    :return: xi_2_H
    :rtype: complex
    """    

    var_alpha = qc.var(alpha)
    mean_beta = qc.expval(beta)
    mean_gamma = qc.expval(gamma)
    return (2*var_alpha)/np.sqrt(mean_beta**2+mean_gamma**2) 

def get_xi_2_S(qc,return_n0=False):
    """

    :param qc: circuit object
    :type qc: Circuit
    :param return_n0: if return n0. Defaults to False., defaults to False
    :type return_n0: bool, optional
    :return: xi_2_S
    :rtype: complex
    """    

    use_gpu = qc.use_gpu
    if use_gpu:
        N = qc.N
        mean_x = qc.expval(type="x")
        mean_y = qc.expval(type="y")
        mean_z = qc.expval(type="z")
        mag_mean_J = torch.sqrt(torch.real(mean_x**2+mean_y**2+mean_z**2))
        theta = torch.arccos(mean_z/mag_mean_J)
        mean_y = torch.real(mean_y)
        if mean_y > 0:
            phi = torch.arccos(mean_x/(mag_mean_J*torch.sin(theta)))
        else:
            phi = 2*torch.pi - torch.arccos(mean_x/(mag_mean_J*torch.sin(theta)))
        if theta == 0 or isclose(theta,np.pi):
            phi = torch.tensor(0)
        n1 = torch.tensor([-torch.sin(phi),torch.cos(phi),0]).type(torch.complex128)
        n2 = torch.tensor([torch.cos(theta)*torch.cos(phi),0,-torch.sin(theta)]).type(torch.complex128)
        cov = qc.expval(type="cov",use_vector=True,n1=n1,n2=n2)/2
        # *(1e-16)
        mean_n1n2_minus = qc.expval(type="n1n2_minus",use_vector=True,n1=n1,n2=n2)
        mean_n1n2_plus = qc.expval(type="n1n2_plus",use_vector=True,n1=n1,n2=n2)
        xi_2_S_1 =  torch.real(2/N*(mean_n1n2_plus+torch.sqrt(mean_n1n2_minus**2+4*cov**2)))
        xi_2_S_2 =  torch.real(2/N*(mean_n1n2_plus-torch.sqrt(mean_n1n2_minus**2+4*cov**2)))

        if return_n0:
            n0 = torch.tensor([torch.sin(theta)*torch.cos(phi),torch.sin(theta)*torch.sin(phi),torch.cos(theta)]).to(qc.device)
            if xi_2_S_1 < 0 and xi_2_S_2 > 0:
                    return n0,xi_2_S_2
            elif xi_2_S_1 > 0 and xi_2_S_2 < 0:
                return n0,xi_2_S_1
            else:
                return n0,min(xi_2_S_1,xi_2_S_2)
        else:
            if xi_2_S_1 < 0 and xi_2_S_2 > 0:
                return xi_2_S_2
            elif xi_2_S_1 > 0 and xi_2_S_2 < 0:
                return xi_2_S_1
            else:
                return min(xi_2_S_1,xi_2_S_2)
    else: 
        N = qc.N
        mean_x = qc.expval(type="x")
        mean_y = qc.expval(type="y")
        mean_z = qc.expval(type="z")
        mag_mean_J = np.sqrt(np.real(mean_x**2+mean_y**2+mean_z**2))
        theta = np.arccos(mean_z/mag_mean_J)
        if mean_y > 0:
            phi = np.arccos(mean_x/(mag_mean_J*np.sin(theta)))
        else:
            phi = 2*np.pi - np.arccos(mean_x/(mag_mean_J*np.sin(theta)))
        if theta == 0 or isclose(theta,np.pi):
            phi = 0
        n1 = np.asarray([-np.sin(phi),np.cos(phi),0]).astype(np.complex128)
        n2 = np.asarray([np.cos(theta)*np.cos(phi),0,-np.sin(theta)]).astype(np.complex128)
        cov = qc.expval(type="cov",use_vector=True,n1=n1,n2=n2)/2
        # *(1e-16)
        mean_n1n2_minus = qc.expval(type="n1n2_minus",use_vector=True,n1=n1,n2=n2)
        mean_n1n2_plus = qc.expval(type="n1n2_plus",use_vector=True,n1=n1,n2=n2)
        xi_2_S_1 =  2/N*(mean_n1n2_plus+np.sqrt(mean_n1n2_minus**2+4*cov**2))
        xi_2_S_2 =  2/N*(mean_n1n2_plus-np.sqrt(mean_n1n2_minus**2+4*cov**2))
        if return_n0:
            n0 = np.asarray([np.sin(theta)*np.cos(phi),np.sin(theta)*np.sin(phi),np.cos(theta)])
            if xi_2_S_1 < 0 and xi_2_S_2 > 0:
                    return n0,xi_2_S_2
            elif xi_2_S_1 > 0 and xi_2_S_2 < 0:
                return n0,xi_2_S_1
            else:
                return n0,min(xi_2_S_1,xi_2_S_2)
        else:
            if xi_2_S_1 < 0 and xi_2_S_2 > 0:
                return xi_2_S_2
            elif xi_2_S_1 > 0 and xi_2_S_2 < 0:
                return xi_2_S_1
            else:
                return min(xi_2_S_1,xi_2_S_2)

def get_xi_2_R(qc):
    """

    :param qc: circuit object
    :type qc: Circuit
    :return: xi_2_R
    :rtype: complex
    """    
    use_gpu = qc.use_gpu
    if use_gpu:
        n0,xi_2_S = get_xi_2_S(qc,return_n0=True,use_gpu=True)
    else:
        n0,xi_2_S = get_xi_2_S(qc,return_n0=True)
    N = qc.N
    mean_J = qc.expval(type="xyz",use_vector=True,n=n0)
    if use_gpu:
        return (N**2/(4*torch.abs(mean_J)**2))*xi_2_S
    return (N**2/(4*np.abs(mean_J)**2))*xi_2_S

def get_xi_2_D(qc,n):
    """

    :param qc: circuit object
    :type qc: Circuit
    :param n: vector n 
    :type n: list, ndarray, tensor
    :return: xi_2_D
    :rtype: complex
    """    
    N = qc.N
    var = qc.var(type="xyz",use_vector=True,n=n)
    mean = qc.expval(type="xyz",use_vector=True,n=n)   
    return (N*var)/(N**2/4-mean**2)

def get_xi_2_E(qc,n):
    """
    :param qc: circuit object
    :type qc: Circuit
    :param n: vector n 
    :type n: list,ndarray,tensor
    :return: xi_2_E
    :rtype: complex
    """    
    N = qc.N
    var_Jvec = qc.var(type="xyz",use_vector=True,n=n)
    mean_Jvec = qc.expval(type="xyz",use_vector=True,n=n)
    mean_J_2 = qc.expval(type="J2")
    return (N*var_Jvec)/(mean_J_2-N/2-mean_Jvec**2)

def get_xi_2_F(qc,n1,n2,n3):
    """

    :param qc: circuit object
    :type qc: Circuit
    :param n1: vector n1 
    :type n1: list,ndarray,tensor
    :param n2: vector n2
    :type n2: list,ndarray,tensor
    :param n3: vector n3
    :type n3: list,ndarray,tensor
    :return: xi_2_F
    :rtype: complex
    """    
    N=qc.N
    var_J_n1 = qc.var(type="xyz",use_vector=True,n=n1)
    mean_J_n2 = qc.expval(type="xyz",use_vector=True,n=n2)
    mean_J_n3 = qc.expval(type="xyz",use_vector=True,n=n3)
    return (N*var_J_n1)/(mean_J_n2**2+mean_J_n3**2)







"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: 
>>> all rights reserved
________________________________
"""


from tqix.vqa.bounds import *
from tqix.vqa.circuits import *
from tqix.vqa.constants import *
from tqix.vqa.entanglement import *
from tqix.vqa.fitting import *
from tqix.vqa.kraus import *
from tqix.vqa.plateau import *
from tqix.vqa.vqent import *
from tqix.vqa.vqm import *